// -*- c++ -*-
// $Id$
/*
 * Copyright 2005 Spencer Olson
 *
 * $Log$
 *
 */

/** \file
 * A stream of gas flows past a fixed plate at zero incidence angle.  This
 * example demonstrates various types of walls:  specularly reflecting,
 * diffusely reflecting, and absorbing.
 */

#include <vector>
#include <iostream>
#include <fstream>
#include <string>

#include <olson-tools/strutil.h>
#include <olson-tools/physical.h>
#include <olson-tools/bfield.h>
#include <olson-tools/random/random.h>
#include <olson-tools/listutil.h>
#include <olson-tools/KeyedBin.h>
#include <olson-tools/GenericBin.h>

#include <dsmc/DSMC.h>
#include <dsmc/ParticleSrc.h>
#include <dsmc/Wall.h>

typedef DSMC< RK2Integrator, octree::CMAndVolNotShrink > DSMC_ALG;
//typedef DSMC< RK2Integrator, octree::CMAndVolShrink > DSMC_ALG;
//typedef DSMC< RK2Integrator, octree::GeometricDivision > DSMC_ALG;


#include <stdexcept>
#include <sys/times.h>
#include <unistd.h>
#include <signal.h> // so we can issue ^C to gently stop the loading

static bool quit_requested = false;
void quitRequested(int inSignal) {
    quit_requested = true;
}

using namespace physical::constants;
using namespace physical::units;

const double Lp = 0.5*m;                /* thin plate length. */
const double theta = 0.0 / 180.0 * pi;  /* angle to y-axis(?) */
const double dx = Lp * fabs(sin(theta));
const double dy = Lp * fabs(cos(theta));
const double nx = cos(theta);
const double ny = sin(theta);

const double n1[3] = { nx, ny, 0};      /* plate normal */
const double d[3]  = { dx, dy, 0};      /* plate side-length (projected) */
const double o1[3] = { 1.5*m, 0, 0};    /* plate center */

const double v_stream = 40.0 *m/s;

const double input_wall  = -1.5*m;
const double output_wall =  1.5*m;

const double maxx_wall    =  1.5*m;
const double minx_wall    = -1.5*m;

const double T_sys = 273.0 * K; /* system temperature :  all boundaries and initial conditions. */


const Vector<double,3> vol_dx = V3( (maxx_wall-minx_wall),
                                    (output_wall-input_wall),
                                    0.0);

const Vector<double,3> half_vol_dx = 0.5 * vol_dx;


typedef octree::WallBoundaryObject<Particle,octree::MomentumTransferMonitor> plate_type;
typedef octree::WallBoundaryObject<Particle> wall_type;
typedef octree::VacuumBoundaryObject<Particle> vac_type;
typedef octree::IntersectBoundaryObject<Particle> corner_type;

void flatderivs(const double p[VZ+1], const double * time, const double * dt, double rkF[VZ+1], void * ignore);

KeyedBin<int,double,14,100*100> splineBin(0, 100*100-1);
KeyedBin<int,double,14,100*100> tmpSBin(0, 100*100-1);

template <unsigned int L, class Interpolator>
void sh_spline(const double pos[3],
               Vector<double,L> & retval,
               const std::vector<int> & eltv,
               Interpolator * inp);

class StreamVDistribs {
  public:
    typedef GaussianDistrib     P_vx;
    typedef MaxwellianDistrib2D P_vy;
    typedef GaussianDistrib     P_vz;
};

int main(int argc, char *argv[]) {
    /* need to be able to add particles to the system. */
    int8_t particle_type = getParticleTypeIndex("Ar");

    /** The thin flat plate.
     * Note that the normal does not have to be "normalized". */
    const int num_plates = 1;
    plate_type plates[num_plates] = {
        /*        normal   point   v_wall side-length  scaling temp  roughness  allow_movement     nbins  */
        plate_type(V3C(n1), V3C(o1), 0., V3C(d),         m,   T_sys, .9, V3b(true,true,false), 10000),
    };

    /** The parallel diffusely reflecting walls that mimic the surrounding
     * gas.
     * Note that the normal does not have to be "normalized". */
    const int num_walls = 2;
    wall_type walls[num_walls] = {
        /*        normal   point   v_wall side-length  scaling temp  roughness  allow_movement     nbins  */
        /* diffusely reflecting walls for the x-planes. */
        wall_type(V3(-1,0,0), V3(maxx_wall,0,0), V3(0,v_stream,0), V3(0,DBL_MAX,0), m, T_sys,-.9,  V3b(true,true,false), 10000),
        wall_type(V3( 1,0,0), V3(minx_wall,0,0), V3(0,v_stream,0), V3(0,DBL_MAX,0), m, T_sys,1.9,  V3b(true,true,false), 10000)
    };

    /** Vacuum boundary conditions. */
    const int num_vacuo = 2;
    vac_type vacuo[num_vacuo] = {
        /*        normal     point            side-length    scaling */
        /* +- y planes */
        vac_type(V3(0,-1,0), V3(0,output_wall,0), V3(DBL_MAX,0,0), m),
        vac_type(V3(0, 1,0), V3(0,input_wall,0), V3(DBL_MAX,0,0), m)
    };
    /* let's list the flux out of each of the end surfaces -- debug flux */
    vacuo[0].keepRemoved = true;
    vacuo[1].keepRemoved = true;

    GenericBin<double,1000> * v_output_bin[num_vacuo] = {
        new GenericBin<double,1000>(-600*m/s, 600*m/s),
        new GenericBin<double,1000>(-600*m/s, 600*m/s)
    };


    const int num_corners = 4;
    corner_type corners[num_corners] = {
        /*        point   side-length */
        /* diffusely reflecting walls for the x-planes. */
        corner_type(V3(maxx_wall,output_wall,0)),
        corner_type(V3(minx_wall,output_wall,0)),
        corner_type(V3(maxx_wall, input_wall,0)),
        corner_type(V3(minx_wall, input_wall,0))
    };

    /* boundaries to corners. */
    corners[0].boundaries.push_back(&walls[0]);
    corners[0].boundaries.push_back(&vacuo[0]);
    corners[1].boundaries.push_back(&walls[1]);
    corners[1].boundaries.push_back(&vacuo[0]);
    corners[2].boundaries.push_back(&walls[0]);
    corners[2].boundaries.push_back(&vacuo[1]);
    corners[3].boundaries.push_back(&walls[1]);
    corners[3].boundaries.push_back(&vacuo[1]);


    const double Kn = 0.1;
    /* We'll ensure that dt ~= min(0.1 Coll.Time,0.1 1m/v_avg)  */
    const double dt_max =
        0.1 /* we need to allow excursions 1/10 the size of L */
        * 1.0 * m /* characteristic length. */
        / (  v_stream
           + sqrt( 8.0*K_B * T_sys / pi / particle_info[particle_type].mass)
          );
    const double dt_Kn =
        0.1 /* we need to allow excursions 1/10 the size of mean free path */
        * (Kn * 1.0*m) /* mean free path. */
        / sqrt( 8.0*K_B * T_sys / pi / particle_info[particle_type].mass);
    const double dt = std::min(dt_Kn, dt_max);
    double t_max = 20.*s + dt /* dt * 200.0 */;
    int spit_out_data = 0;

    const double desired_samples = 1e5;
    const double sample_wait_time = 2.0*s; /* time before real sampling happens. */
    int sample_interval = int( (t_max-sample_wait_time) / ( desired_samples * dt ) );
    int write_interval = 100 * sample_interval;
    int write_particles_interval = write_interval; /* must be multiple of write_interval */

    std::cout << "(" << getpid() << ")  t_max address:  " << (&t_max) << std::endl;
    std::cout << "(" << getpid() << ")  spit_out_data address:  " << (&spit_out_data) << std::endl;
    std::cout << "(" << getpid() << ")  sample_interval address:  " << (&sample_interval) << std::endl;
    std::cout << "(" << getpid() << ")  write_interval address:  " << (&write_interval) << std::endl;


    /* It is assumed that n_part_per_domain  particles are also spread out
     * over the one meter length * of the 3rd dimension. */
    const double n              = 1.0 / ( M_SQRT2 * particle_info[particle_type].cross_section * Lp * Kn );
    const double domain_area    = vol_dx[X] * vol_dx[Y]; /* this will have to coincide with the wall definitions. */
    const double n_part_per_domain = (domain_area * 1.0*m) * n;
    const double flux_per_area  = (n * v_stream);
    const double avg_flux       = (n * v_stream) * (1.0*m * vol_dx[X]);
    particle_info[particle_type].size   = n_part_per_domain * 2e-6;

    std::cout << "(" << getpid() << ")  v_stream:  " << v_stream << std::endl;
    std::cout << "(" << getpid() << ")  n:  " << n << std::endl;
    std::cout << "(" << getpid() << ")  domain_area:  " << domain_area << std::endl;
    std::cout << "(" << getpid() << ")  n_part_per_domain:  " << n_part_per_domain << std::endl;
    std::cout << "(" << getpid() << ")  flux_per_area:  " << flux_per_area << std::endl;
    std::cout << "(" << getpid() << ")  avg_flux:  " << avg_flux << std::endl;
    std::cout << "(" << getpid() << ")  particle size:  " << particle_info[particle_type].size << std::endl;



    ParticleSrc<FlatParticleDistrib> initatomsrc;
    initatomsrc.seed.type           = particle_type;
    initatomsrc.seed.dt             =  10*us;

    initatomsrc.pos_distrib.delta_x = 0.999999*half_vol_dx;
    initatomsrc.origin              = V3( maxx_wall - half_vol_dx[X],
                                          output_wall - half_vol_dx[Y],
                                          0.0);
    initatomsrc.velocity            = V3( 0.0,  v_stream, 0.0);
    initatomsrc.temperature         = T_sys;
    initatomsrc.pulsed_source       = true;
    initatomsrc.pulse_duration      = dt;
    initatomsrc.flux                = n_part_per_domain/dt;
    initatomsrc.initialize();

    double beta                 = 0.5 * particle_info[particle_type].mass
                                / (K_B * initatomsrc.temperature[Y]);
    double f1                   = fabs(upstream_fraction(v_stream,beta));
    double f2                   =    downstream_fraction(v_stream,beta);

    std::cout << "(" << getpid() << ")  upstream flux fraction:  " << f1 << std::endl;
    std::cout << "(" << getpid() << ")  downstream flux fraction:  " << f2 << std::endl;

    ParticleSrc<FlatParticleDistrib,StreamVDistribs> upatomsrc;
    upatomsrc.seed.type           = particle_type;
    upatomsrc.seed.dt             =  10*us;
    upatomsrc.pos_distrib.delta_x = V3( 0.999999*half_vol_dx[X], 0.0,0.0);
    upatomsrc.origin              = V3( 0.0, 0.999999*output_wall,0.0);
    upatomsrc.v_range.min[Y]      = -5.0;
    upatomsrc.v_range.max[Y]      =  0.0;
    upatomsrc.velocity            = initatomsrc.velocity;
    upatomsrc.temperature         = initatomsrc.temperature;
    upatomsrc.pulsed_source       = false;
    upatomsrc.flux                = (v_stream > 0.0 ? f1:f2) * avg_flux;
    upatomsrc.initialize();

    ParticleSrc<FlatParticleDistrib,StreamVDistribs> dnatomsrc;
    dnatomsrc.seed.type           = particle_type;
    dnatomsrc.seed.dt             =  10*us;
    dnatomsrc.pos_distrib.delta_x = V3( 0.999999*half_vol_dx[X], 0.0,0.0);
    dnatomsrc.origin              = V3( 0.0, 0.999999*input_wall,0.0);
    dnatomsrc.v_range.min[Y]      = 0.0;
    dnatomsrc.v_range.max[Y]      = 5.0;
    dnatomsrc.velocity            = initatomsrc.velocity;
    dnatomsrc.temperature         = initatomsrc.temperature;
    dnatomsrc.pulsed_source       = false;
    dnatomsrc.flux                = (v_stream > 0.0 ? f2:f1) * avg_flux;
    dnatomsrc.initialize();

    DSMC_ALG dsmc;
    dsmc.integrator.derivs      = flatderivs;
    dsmc.allow_collisions       = true;

    for (int i = 0; i < num_corners; i++) {
        dsmc.addBoundaryObject(&corners[i]);
    }

    for (int i = 0; i < num_plates; i++) {
        dsmc.addBoundaryObject(&plates[i]);
    }

    for (int i = 0; i < num_walls; i++) {
        dsmc.addBoundaryObject(&walls[i]);
    }

    for (int i = 0; i < num_vacuo; i++) {
        dsmc.addBoundaryObject(&vacuo[i]);
    }

    dsmc.data_radius_of_influence = 5;

    std::ofstream v_output_bin_fout[num_vacuo];
    v_output_bin_fout[0].open("v_0_bin.dat");
    v_output_bin_fout[1].open("v_1_bin.dat");

    /* ** plotting all particles. */
    std::ofstream fout("particles.plt");
    std::ofstream fdata("data.plt");
    std::ofstream fpout("data.dat");

    std::ofstream fcout("collision-totals.dat");
    fout << "set xlabel 'x'\nset ylabel 'y'\nset zlabel 'z'\n";
    std::string tree_plotopt = "w lp ps 1 t ''" ;
    std::string part_plotopt = "w p pt 6 ps 2 lt 2 t 'particles'";
    std::ostringstream ostr;


    fcout << "#  Collisions Totals\n"
             "#  Sim. Time(s)        inst. coll./s"
             "          avg. coll./s"
             "          avg. coll. selection rate (per atom)"
             "          Num. Super Particles"
             "          Cycle CPU Time(s)\n";

    std::ofstream splinebinout("data_spline.dat");

    std::ofstream dPdtout("dPdt_plate.dat");

    std::ofstream fluxout("fluxout.dat");

    double max_dx[3] = {0.0, 0.0, 0.0};

    /* now let the dsmc have some fun with the particles. */
    try {
        signal(SIGINT, quitRequested);
        double seconds_per_clock_tick = 1.0 / sysconf(_SC_CLK_TCK);

        dsmc.addParticles(initatomsrc.createParticles(dt));

        while ( !quit_requested && dsmc.getTotalTime() <= t_max) {
            dsmc.addParticles(upatomsrc.createParticles(dt));
            dsmc.addParticles(dnatomsrc.createParticles(dt));

            struct tms ti, tf;
            times(&ti);
            dsmc.evolve(dt, NULL);
            times(&tf);

            if (dsmc.getTotalTime() >= sample_wait_time) spit_out_data = 1;

            /* **** TAKE CARE OF SAVING OUTPUT INFORMATION AND EXPUNGE BOUNDARIES. */
            {
                fluxout << dsmc.getTotalTime();

                for (int i = 0; i < num_vacuo; i++ ) {
                    Particle * tmp = vacuo[i].removed;
                    while (tmp) {
                        v_output_bin[i]->bin( tmp->vel[Y] );
                        tmp = tmp->next;
                    }
                    fluxout << '\t' << deleteList<Particle>(vacuo[i].removed);
                    vacuo[i].removed = NULL;
                }

                fluxout << std::endl;
            }

            {
                double cycle_cpu_time = 
                     ( ( (tf.tms_utime + tf.tms_stime) - (ti.tms_utime + ti.tms_stime) )
                       * seconds_per_clock_tick
                     );

                /* plot collision totals every time-step. */
                if (dsmc.numberParticles() > 0) {
                    fcout << dsmc.getTotalTime()       << '\t'
                          << (dsmc.getCycleCollisions()/(dt*dsmc.numberParticles())) << '\t'
                          << (dsmc.getTotalCollisions()/(dsmc.getTotalTime()*dsmc.numberParticles())) << '\t'
                          /* the following is supposed to be collision selection rate
                           * per atom.  That is why the factor of 2 precedes the line.
                           */
                          << (2*dsmc.getTotalCollisionSelections()/(dsmc.getTotalTime()*dsmc.numberParticles())) << '\t'
                          << dsmc.numberParticles() << '\t';
                } else {
                    fcout << dsmc.getTotalTime()
                          << "\t0\t0\t0\t0\t";
                }
                fcout << cycle_cpu_time << std::endl;
            }

            if ((int(dsmc.getTotalTime() / dt) - 1) % sample_interval  == 0) {
                /* instead of just writing out the data, we'll bin it and
                 * then write out the binned data.
                 */
                const int ELT_LEN=11;
                const int ELT[ELT_LEN] = {
/* 6 */             DSMCData<>::NUMBER_DENSITY,
/* 7 */             DSMCData<>::NUMBER,
/* 8 */             DSMCData<>::MAX_SIGMA_RELSPEED,
/* 9 */             DSMCData<>::NUMBER_COLLISION_PAIRS_REMAINDER,
/* 10*/             DSMCData<>::NUMBER_COLLISIONS,
/* 11*/             DSMCData<>::AVG_V_X,
/* 12*/             DSMCData<>::AVG_V_Y,
/* 13*/             DSMCData<>::AVG_V_Z,
/* 14*/             DSMCData<>::AVG_V2_X,
/* 15*/             DSMCData<>::AVG_V2_Y,
/* 16*/             DSMCData<>::AVG_V2_Z
                };
                const std::vector<int> eltv(ELT, ELT+ELT_LEN);

                int row_len = (int)sqrt(splineBin.nBins());
                const double dx = vol_dx[X] / ((double)row_len - 1.0);
                for (unsigned int i = 0; i < splineBin.nBins(); i++) {
                    Vector<double,14> v;
                    Vector<double,11> & v11 = *((Vector<double,11>*)(&v.val[3]));
                    Vector<double,3> x = V3C(v.val);

                    /* we'll assume that vol_dx[X] >= vol_dx[Y] */
                    x[X] = (i / row_len) * dx - half_vol_dx[X];
                    x[Y] = (i % row_len) * dx - half_vol_dx[X];
                    x[Z] = 0.0;

                    v11 = 0.0;
                    sh_spline(x.val, v11, eltv, dsmc.getInterpolator());

                    v[0] = x[X]; v[1] = x[Y]; v[2] = x[Z];
                    splineBin.bin(i, v);
                }
            }

            /* **** TAKE CARE OF PLOTTING ALL PARTICLES. */
            if (dsmc.getTotalTime() >= (t_max-dt) || (int(dsmc.getTotalTime() / dt) - 1) % write_interval  == 0) {
                std::string tot_time = to_string(dsmc.getTotalTime());
#if 0
                if (int(dsmc.getTotalTime() / dt - 1) % 100 == 0) {
                /* note that this octree is not necessarily representative of the
                 * current particles.  This is because there has been a call to
                 * dsmc.evolve since the Octree was built.
                 */
                dsmc.plotParticlesAndTree(fout,
                                          tot_time + " s",
                                          tree_plotopt,
                                          part_plotopt)
                    << "pause -1\n" << std::flush;
                }
#endif

                /* now write out the momentum transfer in the plate, then zero
                 * the monitor. */
                plates[0].monitor.P /= dt * write_interval;
                dPdtout << tot_time << '\t' << plates[0].monitor.P << std::endl;
                plates[0].monitor.P = 0.0;

                if (dsmc.getInterpolator()) {
                    /* ok, let's spit out the data tree. */
#if 0
                    dsmc.getInterpolator()->
                         plotParticlesAndTree(fdata,
                                              "DSMC Data Tree:  " + tot_time + " s",
                                              tree_plotopt,
                                              part_plotopt)
                     << "pause -1\n" << std::flush;
#endif

#if 0
                    fpout << "#  Data Particles\n#  Time:  " << tot_time << " s\n";
                    dsmc.getInterpolator()->printParticles(fpout) << "\n\n";

                    /* now let's zero the collisions-per-octree-node counters
                     * so that we only get number collisions per sample time.
                     */
                    DSMCData<> ** d = dsmc.getInterpolator()->getParticles();
                    for (int i = dsmc.getInterpolator()->numberParticles() - 1; i >= 0; i--) {
                        d[i]->data.name.collisions = 0;
                    }
#endif

                    /* write out spline data. */
                    tmpSBin = splineBin;
                    /* take average. */
                    for (unsigned int i = 0; i < tmpSBin.nBins(); i++) {
                        if (tmpSBin.hist[i]) tmpSBin.bins[i] /= tmpSBin.hist[i];
                    }

                    /* write out new shmooshed together spline data. */
                    splinebinout << "#  Averaged bin across channel\n#  Time:  " << tot_time << " s\n";
                    tmpSBin.print(splinebinout,"") << "\n\n" << std::flush;

                    if ( !spit_out_data ) {
                        /* we'll just keep clearing the bins until we need to
                         * really start time-averaging.
                         */
                        splineBin.clearBins();
                    }
                    /* now let's zero the collisions-per-octree-node counters
                     * so that we only get number collisions per sample time.
                     */
                    DSMCData<> ** d = dsmc.getInterpolator()->getParticles();
                    for (int i = dsmc.getInterpolator()->numberParticles() - 1; i >= 0; i--) {
                        d[i]->data.name.collisions = 0;
                    }
                } else if (dsmc.numberParticles() > 0 ) {
                    /* we have to make our own interpolator since DSMC didn't need
                     * to construct it. */
                    octree::OctreeInterpolator<DSMCData<>,Particle> ointerp;
                    DSMCData<>::InitInfo init = {dsmc.getParticles()};

                    ointerp.createDataOctree(dsmc.getOctree(), init);
                    ointerp.setRadiiOfInfluence(dsmc.data_radius_of_influence);

                    fpout << "#  Data Particles\n#  Time:  " << tot_time << " s\n";
                    ointerp.printParticles(fpout) << "\n\n";
                }

            }/* if */

            for (int i = X; i <= Z; i++)
                max_dx[i] = std::max(max_dx[i], dsmc.max_particle_transit(i));
        }/* while */
    } catch (std::exception & e) {
        std::cout << "exception:  " << e.what() << std::endl;
    }

    for (int i = 0; i < num_vacuo; i++) {
        v_output_bin[i]->print(v_output_bin_fout[i],"");
        delete v_output_bin[i];
        v_output_bin_fout[i].close();
    }

    fout.close();
    fpout.close();
    fdata.close();
    fcout.close();

    dPdtout.close();

    std::cout << "max Particle Transit:  ("
              << max_dx[X] << ','
              << max_dx[Y] << ','
              << max_dx[Z] << ')'
              << std::endl;

    return 0;
}

void flatderivs(const double p[VZ+1], const double * time, const double * dt, double rkF[VZ+1], void * ignore) {
    /* a 2d problem. */
    rkF[X] = p[VX];
    rkF[Y] = p[VY];
    rkF[Z] = 0.0;

    /* no forces. */
    rkF[VX] = 0.0;
    rkF[VY] = 0.0;
    rkF[VZ] = 0.0;
}

template <unsigned int L, class Interpolator>
void sh_spline(const double pos[3],
               Vector<double,L> & retval,
               const std::vector<int> &eltv,
               Interpolator * inp) {
    if (!inp) return;
    if (L != eltv.size()) {
        throw std::runtime_error("elt vector != size of return value!");
    }

    std::vector<double> interpolant = inp->interpolate(pos, eltv);

    if (interpolant.size() > 0) {
        for (unsigned int i = 0; i < L; i++) {
            retval[i] = interpolant[i];
        }
    } else {
        for (unsigned int i = 0; i < L; retval[i++] = 0.0);
    }
}

