// -*- c++ -*-
// $Id$
/** \file
 * Simple test function for building an Octree of N particles.
 * Copyright 2004 Spencer Olson
 *
 * $Log$
 *
 */

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <sstream>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <olson-tools/random/random.h>
#include <olson-tools/listutil.h>
#include <sys/times.h>
#include <unistd.h>

#include <dsmc/Particle.h>
#include <dsmc/OctreeContainer.h>

#define BOX_SIDE 100

void subtime(struct timeval &ti, struct timeval &tf, struct timeval &result) {
    if (tf.tv_usec < ti.tv_usec) {
        tf.tv_usec += 1000000;
        tf.tv_sec--;
    }
    result.tv_sec = tf.tv_sec - ti.tv_sec;
    result.tv_usec = tf.tv_usec - ti.tv_usec;
}

static double seconds_per_clock_tick = 1.0 / sysconf(_SC_CLK_TCK);

//typedef octree::CMAndVolShrink CMVolMeasure;
typedef octree::CMAndVolNotShrink CMVolMeasure;
//typedef octree::GeometricDivision CMVolMeasure;

typedef octree::OctreeContainer<Particle,octree::nothing,CMVolMeasure> Container;

inline void run(const int N, const int Nmin, std::ostream & ftiming);

int main() {
    Container::OctreeType::freetoheap(6000000); /* about 100MB */

    Particle::freetoheap(6000000);

    std::ofstream ftiming ("octree-timing.dat");

#ifdef DSMC_USE_PTHREAD
#  define PTHREAD_COLUMN        "        Num. Threads        "
#else
#  define PTHREAD_COLUMN        ""
#endif

    ftiming << "# " PTHREAD_COLUMN "Number particles        Min Part/Octree Leaf        gettimeofday() diff        times(...) diff\n";

    try {
        while (true) {
            int N = -10, Nmin = -1;

#ifdef DSMC_USE_PTHREAD
            int nThreads = 1;
            std::cout << "Enter number of threads to use:  ";
            std::cin >> nThreads;
#endif

            std::cout << "Enter number of particles to include in build:  ";
            std::cin >> N;
            std::cout << "Enter rough number of particles per Octree node:  ";
            std::cin >> Nmin;
#ifndef STATIC_SEED
#  warning "Define -DSTATIC_SEED=<seed> to seed the random number generator"
#endif
#ifdef STATIC_SEED
            uint32 seed = STATIC_SEED;
            MTRNGseedV1(&seed);
#endif
            if (N <= 0 || Nmin <= 0) exit(1);

#ifdef DSMC_USE_PTHREAD
            if (nThreads <= 0) exit(1);
            pthreadLimiter.set_max_pthreads(nThreads);
#endif

            run(N,Nmin,ftiming);
            ftiming << std::flush;

            std::cout << "Build complete.  "
                         "Press CTRL-C to avoid overwriting tree-detail output files."
                      << std::endl;
        }
    } catch (std::exception & e) {
        std::cout << "exception:  " << e.what() << std::endl;
    }

    ftiming.close();

    return 0;
}

inline void run(const int N, const int Nmin, std::ostream & ftiming) {
    /* do it twice to ensure full allocation of Octree nodes, so that
     * time won't be wasted allocating new memory from the system.
     */
    Particle * particles = NULL;
    Particle** pPtrPtr = &particles;

    /* allocate and initialize all particles */
    for( int i = N; i > 0; i--) {
        (*pPtrPtr) = new Particle();
        (*pPtrPtr)->pos[X] = MTRNGrand() * BOX_SIDE;
        (*pPtrPtr)->pos[Y] = MTRNGrand() * BOX_SIDE;
        (*pPtrPtr)->pos[Z] = MTRNGrand() * BOX_SIDE;
        pPtrPtr = &((*pPtrPtr)->next);
    }

    Container container(Container::TEMPORARY);
    
    for (int j = 0; j < 2; j++) {
        container.addParticles(particles);

        /* build the octree via quicksort */
        //std::cerr << "starting sort"<<std::endl<<std::flush;
        struct timeval tv[3];
        struct tms ti, tf;

        std::cout << "Starting " << j << "th iteration" << std::endl;
        times(&ti);
        gettimeofday(&tv[0],NULL);
        container.buildOctree(Nmin);
        gettimeofday(&tv[1],NULL);
        times(&tf);
        std::cout << "Finished" << std::endl;
        subtime(tv[0],tv[1],tv[2]);

        double t = tv[2].tv_sec + double(tv[2].tv_usec)*1e-6L;
        double cycle_cpu_time = 
             ( ( (tf.tms_utime + tf.tms_stime) - (ti.tms_utime + ti.tms_stime) )
               * seconds_per_clock_tick
             );

        if (j == 1) {
            /* record only the second iteration. */

#ifdef DSMC_USE_PTHREAD
            ftiming << pthreadLimiter.get_max_pthreads() << '\t';
#endif
            ftiming << N << '\t' << Nmin << '\t' << t << '\t' << cycle_cpu_time << std::endl;
        }

        /* Now we write out some files to describe the generated tree in
         * detail.  Warning:  This can be big. */
#ifndef WRITE_DETAILS
#  warning "Define WRITE_DETAILS to get the tree and particles written"
#endif
#ifdef WRITE_DETAILS
        {
            double PX1  = 100*double(MTRNGrand())/double(RAND_MAX),
                   PY1  = 100*double(MTRNGrand())/double(RAND_MAX),
                   PZ1  = 100*double(MTRNGrand())/double(RAND_MAX);

            Container::OctreeType & otree = *container.getOctree();
            otree.findNearestLeaf(PX1,PY1,PZ1).setMarked(true);

            std::ofstream foutt ("tree-dump.dat");
            foutt << "Octree:\n" << otree << std::endl;

            std::ofstream foutv ("volume.dat");
            /* ****************** BEGIN GNUPLOT OUTPUT ********************** */
            /* plot out the gnuplot plot of all particles and bounding boxes of leaf
             * octree nodes */
            std::string plotcmd;
            std::string plotopt = "w lp ps 1" ;
            std::ostringstream ostr;
            otree.gnuplotLeavesVol(ostr << "#Volume definitions:\n");
            foutv
            	      << "set xlabel 'x'\nset ylabel 'y'\nset zlabel 'z'\nset nokey\n"
            	      << "splot  '-' w p ps 4, '-' using 1:2:3 w p ps 2,"
        	      << Container::OctreeType::createGnuplotLeaveSPLOT(plotcmd, plotopt, otree.leaves)
            	      << '\n'
        	      << PX1 << '\t' << PY1 << '\t' << PZ1 << "\ne\n";

            for (int i = 0; i < container.numberParticles(); i++) { 
        	foutv << *container.getParticles()[i] << '\n';
            }

            foutv     << "e\n"
            	      << ostr.str()
            	      << "pause -1\n" << std::flush;
            /* ******************** END GNUPLOT OUTPUT ********************** */

            foutt.flush();
            foutt.close();

            foutv.flush();
            foutv.close();
        }
#endif

        container.freeParticles();
        container.freeOctree();
    }

    deleteList(particles);
}
