// -*- c++ -*-
// $Id: testtiming.C,v 1.2 2005/06/07 10:00:59 olsonse Exp $
/** \file
 * Simple test function for timing the Octree sort.
 * Copyright 2004 Spencer Olson
 *
 * $Log: testtiming.C,v $
 * Revision 1.2  2005/06/07 10:00:59  olsonse
 * Many changes over the course of time
 *
 * Revision 1.1.1.1  2005/01/08 04:57:26  olsonse
 * Initial CVS import (finally).
 *
 */

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <sstream>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <olson-tools/random/random.h>
#include <olson-tools/listutil.h>
#include <sys/times.h>
#include <unistd.h>

#include <dsmc/Particle.h>
#include <dsmc/OctreeContainer.h>

#define OUTDIR "./"

#define BOX_SIDE 100

using octree::UniformLeafModel;

//#define SHOW_USAGE

#define UNIFORMLEAFMODEL 0
#define DSMCLEAFMODEL 1

#if defined(LEAFMODEL) && (LEAFMODEL == UNIFORMLEAFMODEL)
#  define LEAFMODELCLASS UniformLeafModel
#elif defined(LEAFMODEL) && (LEAFMODEL == DSMCLEAFMODEL)
#  define LEAFMODELCLASS DSMCLeafModel
#else
#  define LEAFMODEL UNIFORMLEAFMODEL
#  define LEAFMODELCLASS UniformLeafModel
#  ifdef SHOW_USAGE
#    warning "Define LEAFMODEL to specify octree leaf model (default: UniformLeafModel)"
#    warning "possibilities:  "
#    warning "    0 (UniformLeafModel)"
#    warning "    1 (DSMCLeafModel)"
#  endif
#endif

void subtime(struct timeval &ti, struct timeval &tf, struct timeval &result) {
    if (tf.tv_usec < ti.tv_usec) {
        tf.tv_usec += 1000000;
        tf.tv_sec--;
    }
    result.tv_sec = tf.tv_sec - ti.tv_sec;
    result.tv_usec = tf.tv_usec - ti.tv_usec;
}

static double seconds_per_clock_tick = 1.0 / sysconf(_SC_CLK_TCK);

#ifdef OCTREE_DIVISION
typedef octree::OCTREE_DIVISION CMVolMeasure;
#else
//typedef octree::CMAndVolShrink CMVolMeasure;
typedef octree::CMAndVolNotShrink CMVolMeasure;
//typedef octree::GeometricDivision CMVolMeasure;
#  ifdef SHOW_USAGE
#    warning "Define OCTREE_DIVISION to specify (defaults to CMAndVolNotShrink)"
#    warning "possibilities:"
#    warning "    CMAndVolShrink"
#    warning "    CMAndVolNotShrink"
#    warning "    GeometricDivision"
#  endif
#endif

typedef octree::OctreeContainer<Particle,octree::nothing,CMVolMeasure,LEAFMODELCLASS> Container;

template <class LeafModel>
inline void run(const LeafModel & leafModel, const int N, std::ostream & fout);

LEAFMODELCLASS leafModel;

int main() {
    const double NMIN  = 1e1,/* number of particles to sort */
                 NMAX  = 6e6,/* number of particles to sort */
                 NstepFraction = 1.5,/* number of particles to sort */
                 Nmin_MIN = 1,
                 Nmin_MAX = 1e6,
                 Nmin_stepFraction = 10;

    Container::OctreeType::freetoheap(6000000); /* about 100MB */

    Particle::freetoheap(6000000);

    std::ofstream fout (OUTDIR"/octree-timing.dat");

    fout << "# Number particles        Min Part/Octree Leaf        gettimeofday() diff        times(...) diff\n";

    try {
        for (double Nmin = Nmin_MIN; Nmin <= Nmin_MAX; Nmin*=Nmin_stepFraction) {
            for (double N = NMIN; N < NMAX; N*=NstepFraction) {
#if (LEAFMODEL == DSMCLEAFMODEL)
#  error "This leaf model is not allowed for yet."
                std::string p_type;
                std::cout << "Enter particle type:  ";
                std::cin >> p_type;

                int p_size = 1;
                std::cout << "Enter particle size:  ";
                std::cin >> p_size;


                int particle_type = getParticleTypeIndex(p_type.c_str());
                particle_info[particle_type].size = p_size;

                leafModel.particle_type = particle_type;
#elif (LEAFMODEL == UNIFORMLEAFMODEL)
                leafModel.max_particles_per_leaf = Nmin;
#else
#  error "unknown leaf model"
#endif
                run(leafModel,int(N),fout);
            }/* for */
            fout << std::endl;
        }/* for */
    } catch (std::exception & e) {
        std::cout << "exception:  " << e.what() << std::endl;
    }

    fout.close();

    return 0;
}

template <class LeafModel>
inline void run(const LeafModel & leafModel, const int N, std::ostream & fout) {
    /* do it twice to ensure full allocation of Octree nodes, so that
     * time won't be wasted allocating new memory from the system.
     */
    Particle * particles = NULL;
    Particle** pPtrPtr = &particles;

    /* allocate and initialize all particles */
    for( int i = N; i > 0; i--) {
        (*pPtrPtr) = new Particle();
        (*pPtrPtr)->pos[X] = MTRNGrand() * BOX_SIDE;
        (*pPtrPtr)->pos[Y] = MTRNGrand() * BOX_SIDE;
        (*pPtrPtr)->pos[Z] = MTRNGrand() * BOX_SIDE;
        pPtrPtr = &((*pPtrPtr)->next);
    }

    Container container(Container::TEMPORARY);
    
    for (int j = 0; j < 2; j++) {
        container.addParticles(particles);

        /* build the octree via quicksort */
        //std::cerr << "starting sort"<<std::endl<<std::flush;
        struct timeval tv[3];
        struct tms ti, tf;

        times(&ti);
        gettimeofday(&tv[0],NULL);
        container.buildOctree(leafModel);
        gettimeofday(&tv[1],NULL);
        times(&tf);
        subtime(tv[0],tv[1],tv[2]);

        double t = tv[2].tv_sec + double(tv[2].tv_usec)*1e-6L;
        double cycle_cpu_time = 
             ( ( (tf.tms_utime + tf.tms_stime) - (ti.tms_utime + ti.tms_stime) )
               * seconds_per_clock_tick
             );

        if (j == 1) {
            /* record only the second iteration. */
            fout << N << '\t' << leafModel << '\t' << t << '\t' << cycle_cpu_time << std::endl;
        }

        container.freeParticles();
        container.freeOctree();
    }

    deleteList(particles);
}
