// -*- c++ -*-
// $Id$
/*
 * Copyright 2005 Spencer Olson
 *
 * $Log$
 *
 */


#ifndef BOUNDARYOBJECT_H
#define BOUNDARYOBJECT_H

#if defined(__INTEL_COMPILER)
#  include <mathimf.h>
#else
#  include <math.h>
#endif

#include <vector>

#include "FlatSurface.h"
#include "Wall.h"
#include "Octree.h"

namespace octree {

/** Description of an interaction with a boundary.
 * - UNCHANGED:       particle is UNeffected by interaction.
 * - CHANGED:         particle IS effected by interaction.
 * - REMOVED:         _Particle needs to be removed from the system.
 *     If keepRemoved == true, then the removed _Particle is placed in the
 *     above "removed" list.
 */
enum INTERACTION {
    UNCHANGED=0,
    CHANGED,
    REMOVED
};

/** Parent boundary object class for use with OctreeContainer and especially
 * with DSMC.
 * This base boundary class takes care of scaling boundaries for use in DSMC.
 * This class furthermore and provides a linked list to which particles can be
 * attached that have been caused to be removed from a system by this
 * boundary.   
 *
 * @param _Particle Define the particle type with which this boundary will be
 * used.  
 */
template <class _Particle>
class BoundaryObject {
  public:
    /** Default constructor for base boundary class. */
    BoundaryObject() :
        volb_orig(), volb_scaled(),
        volb_scale_factor(1.0),
        keepRemoved(false), removed(NULL) {}

    /** Virtual destructor that does nothing. Do we really have to implement
     * this?  */
    virtual ~BoundaryObject() {}

    /** Minimal bounding volume of BoundaryOjbect. */
    vol_bounds volb_orig;

    /** Scaled bounded volume (scaled by scaleVolume). */
    vol_bounds volb_scaled;

    /** Scaling factor for defining volb_scaled. */
    double volb_scale_factor;

    /** Container should put removed particles in the following "removed" list
     * if this is true. */
    bool keepRemoved;

    /** "removed" list. */
    _Particle * removed;

    /** Provide the interaction for this boundary.
     * @param pold
     *     (input) The position of the particle prior to any current timestep
     *     movement (important for determining reflections; (output) The
     *     new 'old' position for the changed particle (such as the location
     *     of a reflection with the new velocity) IF return value is
     *     !UNCHANGED.
     *
     * @param p
     *     (input) The current position of the particle; (output) The new
     *     position of the particle if the return value is !UNCHANGED.
     *
     * @returns the general result of the interaction that occurred;
     * (UNCHANGED, CHANGED, REMOVED) reflects the status of the particle after
     * @see enum INTERACTION.
     */
    virtual enum INTERACTION interact(_Particle & pold, _Particle & p) = 0;

    /** scale the volume extents such that they only have to possibly enclose
     * _Particle/s that have moved a maximum distance of dx.
     * Specializations of BoundaryObject MUST implement interact such that it
     * tests whether the interaction should actually happen.
     * @see interact
     */
    inline void scaleVolume( const double dx[] /* == v_max * dt */) {
        /* copy over all original bounds. */
        volb_scaled = volb_orig;

        /* scale all sides by dx[i]. */
        for (int i = X; i <= Z; i++) {
            volb_scaled.min[i] -= fabs(dx[i]) * volb_scale_factor;
            volb_scaled.max[i] += fabs(dx[i]) * volb_scale_factor;
        }/* for */
    }
};

template<class _Particle>
class FlatBoundaryObject : public virtual FlatSurface, public BoundaryObject<_Particle> {
    typedef  FlatSurface super1;
    typedef BoundaryObject<_Particle> super2;
  public:
    FlatBoundaryObject(const Vector<double,3> & normal = V3(1.0,0.0,0.0),
                       const Vector<double,3> & point = V3(0.0,0.0,0.0),
                       const Vector<bool,3> & allow_movement = true,
                       const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                       const double & length_scaling = physical::unit::meter ) : 
                       super1(), super2(), dl(0.0) {
        setSurface(normal,point,allow_movement,side_length,length_scaling);
    }

    virtual ~FlatBoundaryObject()  {}

    virtual enum INTERACTION interact(_Particle & pold, _Particle & p)  = 0;

    /** Set the surface of the flat boundary.
     * @param normal
     *     Normal to the surface of the wall.
     *
     * @param point
     *     A point lying on the wall somewhere.
     *
     * @param side_length
     *     Vector of lengths of FlatSurface BoundaryObject in each of the 3-D
     *     directions.
     *
     * @param length_scaling
     *     Tiny bit of voodoo:  the (smallest) length scale over which particles are
     *     expected to move within one time frame.
     *
     * @see FlatSurface::setSurface.
     * */
    inline void setSurface(const Vector<double,3> & normal,
                           const Vector<double,3> & point,
                           const Vector<bool,3> & allow_movement,
                           const Vector<double,3> & side_length,
                           const double & length_scaling) {
        super1::setSurface(normal, point, allow_movement);

        dl = side_length;

        /* let's make sure that the side length is only positive.
         * We also ensure that each side is at least 2*(small_number) in length (so that
         * contains doesn't inadvertently fail).
         *
         * The choice of this 1e-5 factor was almost arbitrary and not
         * <i>really</i> well studied.
         */
        double eps = 1e-5*length_scaling;
        dl[X] = std::max(eps, fabs(dl[X]));
        dl[Y] = std::max(eps, fabs(dl[Y]));
        dl[Z] = std::max(eps, fabs(dl[Z]));

        Vector<double,3> & mx = V3C(super2::volb_orig.max);
        Vector<double,3> & mn = V3C(super2::volb_orig.min);

        mx = dl;
        mx *= 0.5;
        mx += point;

        mn = dl;
        mn *= -0.5;
        mn += point;
    }

    inline const Vector<double,3> & size() const { return dl; }

  private:
    Vector<double,3> dl;
};


/** An intersection between two or more boundaries which requires preservation
 * of interaction order. 
 *
 * WARNING:  it may be necessary to set volb_scale_factor to greater than 1.0.
 * For example, consider an intersection of two flat plates with an angle
 * theta.  If theta < (pi/2) or theta > (3 pi/2), then it is necessary to scale the
 * bounding box by an additional factor equal to (1/sin(theta)).  Otherwise,
 * it would be possible for particles, in a DSMC simulation for example, to
 * intersect with the wrong surface.  For an acute (inside) angle, it is necessary to
 * allow multiple wall collisions in order to avoid particles leaking through
 * walls.  For an (outside) angle (theta > (3 pi/2)), it is necessary to
 * preserve collision order, else particles may collide with the wrong wall.
 * 
 * <br>
 * example scale factor for intersecting flat plates with (any) acute angles:
 * <br>
 * volb_scale_factor = 1./sin(theta) <br>
 * where theta is the smallest angle made by the plates.  
 */
template<class _Particle>
class IntersectBoundaryObject : public BoundaryObject<_Particle> {
    typedef BoundaryObject<_Particle> super;

  public:
    IntersectBoundaryObject( const Vector<double,3> & point = V3(0.0,0.0,0.0),
                             const Vector<double,3> & side_length = V3(0.0,0.0,0.0) ) : 
                       super(), boundaries() {
        Vector<double,3> & mx = V3C(super::volb_orig.max);
        Vector<double,3> & mn = V3C(super::volb_orig.min);

        mx = side_length;
        mx *= 0.5;
        mx += point;

        mn = side_length;
        mn *= -0.5;
        mn += point;
    }

    virtual ~IntersectBoundaryObject()  {}

    /** Preserve interaction order among a set of intersecting boundaries.
     */
    virtual enum INTERACTION interact(_Particle & pold, _Particle & p) {
        enum INTERACTION retval = UNCHANGED;

        int blen = boundaries.size();
        do {
            /* first check to see if any boundaries interact with the
             * particles; if so, find the one that should interact first. */
            int i_tmin = -1;
            double t_min = DBL_MAX;

            for (int i = 0; i < blen; i++) {
                FlatBoundaryObject<_Particle> & bo = *boundaries[i];
                double t = bo.intersectTime(pold, p, bo.volb_orig);

                if (t < t_min) {
                    t_min = t;
                    i_tmin = i;
                }
            }

            /* now interact */
            if (t_min <= 0.0) {         /* intersection in the past. */
                FlatBoundaryObject<_Particle> & bo = *boundaries[i_tmin];
                retval = bo.interact(pold, p);

                if (retval == UNCHANGED) {
                    log_severe(
                        "expected particle[(%g,%g,%g) -> (%g,%g,%g)]/"
                        "wall(=%d) at t = %g\n",
                        pold.pos[X], pold.pos[Y], pold.pos[Z],
                        p.pos[X], p.pos[Y], p.pos[Z],
                        i_tmin, t_min
                    );
                    break;
                }
            } else break;

        /* We continue in this while loop until we do not have any more
         * interactions.  */
        } while (retval != REMOVED);

        return retval;
    }

    std::vector< FlatBoundaryObject<_Particle> * > boundaries;
};

/** A null (do nothing) class for monitoring interactions with the wall. */
class NullMonitor {
  public:
    template <class _Particle>
    inline void start(const _Particle & p) {}

    template <class _Particle>
    inline void stop (const _Particle & p) {}
};

/** A class for monitoring momentum transfer from interactions of Particle/s
 * with the wall. */
class MomentumTransferMonitor {
  public:
    inline MomentumTransferMonitor() : P(0.0) {}
    Vector<double,3> P;  /* summary change in momentum */

    template <class _Particle>
    inline void start(const _Particle & p) {
        P.addFraction(  p.mass() * p.size(), V3C(p.vel) );
    }

    template <class _Particle>
    inline void stop(const _Particle & p) {
        P.addFraction( -p.mass() * p.size(), V3C(p.vel) );
    }
};

template <class _Particle, class InteractMonitor = NullMonitor>
class WallBoundaryObject : public virtual FlatBoundaryObject<_Particle>, public Wall {
    typedef FlatSurface super0;
    typedef FlatBoundaryObject<_Particle> super1;
    typedef Wall super2;
  public:
    WallBoundaryObject(const Vector<double,3> & normal = V3(1.0,0.0,0.0),
                       const Vector<double,3> & point = 0.0,
                       const Vector<double,3> & velocity = 0.0,
                       const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                       const double & length_scaling = physical::unit::meter,
                       const double & temp = 300 * physical::unit::Kelvin,
                       const double & rough = -1,
                       const Vector<bool,3> & allow_movement = true,
                       const int & number_distrib_bins = 1000 )
        : super0(), super1(), super2() {
        setSurface(normal,point,velocity,side_length,length_scaling,temp,rough,allow_movement,number_distrib_bins);
    }

    inline void setSurface(const Vector<double,3> & normal,
                           const Vector<double,3> & point,
                           const Vector<double,3> & velocity = 0.0,
                           const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                           const double & length_scaling = physical::unit::meter,
                           const double & temp = 300 * physical::unit::Kelvin,
                           const double & rough = -1,
                           const Vector<bool,3> & allow_movement = true,
                           const int & number_distrib_bins = 1000 ) {
        /* a little redundant:  both super1 and super2 call
         * super0::setSurface.  Oh well. */
        super1::setSurface(normal, point, allow_movement, side_length, length_scaling);
        super2::setSurface(normal, point, velocity, temp, rough, allow_movement, number_distrib_bins);
    }

    virtual ~WallBoundaryObject()  {}

    virtual enum INTERACTION interact(_Particle & pold, _Particle & p) {
        /* have to record the old particle before reflecting because the
         * reflection code changes the 'old' particle upon reflections. */
        monitor.start(pold);
        switch (super2::reflect(pold,p,super1::volb_orig)) {
            case NO_REFLECTION:
                monitor.stop(pold);
                return UNCHANGED;
            default:
                monitor.stop(p);
                return CHANGED;
        }
    }
        
    InteractMonitor monitor;

};

/** Vacuum boundary.
 * Vacuum boundary removes particles whose current position is on the wrong
 * side of the surface (the normal points towards the inside of the vacuum
 * chamber).
 */
template <class _Particle>
class VacuumBoundaryObject : public FlatBoundaryObject<_Particle> {
    typedef FlatSurface super0;
    typedef FlatBoundaryObject<_Particle> super1;
  public:
    /** Shrink the vacuum object to only remove particles within its original
     * bounding volume.  This is also useful if the user also tweaks
     * (properly increases) the bounding volumes so that specific additional
     * regions are included in the vacuum.  Default: false.
     */
    bool shrink;

    VacuumBoundaryObject(const Vector<double,3> & normal = V3(1.0,0.0,0.0),
                         const Vector<double,3> & point = 0.0,
                         const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                         const double & length_scaling = physical::unit::meter,
                         const bool & _shrink = false )
        : super0(), super1(), shrink(_shrink) {
        setSurface(normal,point,side_length,length_scaling);
    }

    inline void setSurface(const Vector<double,3> & normal,
                           const Vector<double,3> & point,
                           const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                           const double & length_scaling = physical::unit::meter ) {
        /* a little redundant:  both super1 and super2 call
         * super0::setSurface.  Oh well. */
        const Vector<bool,3> allow_movement = true;
        super1::setSurface(normal, point, allow_movement, side_length, length_scaling);
    }

    virtual ~VacuumBoundaryObject()  {}

    /** Vacuum boundary interaction.
     * Interaction only depends on current position of particle.
     */
    virtual enum INTERACTION interact(_Particle & pold, _Particle & p) {
        double side_sign = 0;
        intersectTime(p,side_sign);

        if (side_sign <= 0 &&
            (!shrink || super1::volb_orig.contains(p.pos))
           ) {
            return REMOVED;
        } else return UNCHANGED;
    }

};

/** \example testTeleporter.C
 * Creates a simple box container for particles in a DSMC simulation.  The
 * edges of the box are implemented by a TeleporterBoundaryObject.  The
 * TeleporterBoundaryObject offsets the position of any incident particles
 * by a given amount.  
 */

/** \example testWalls.C
 * Demonstrates wall objects, vacuum objects, and corner objects.  
 * This example simulates flow past a thin plate that is implemented using a
 * WallBoundaryObject.  A monitor is placed on the wall to record the total
 * momentum transfer to the wall.  The other walls are either purely specular
 * reflecting surfaces or purely diffusely reflecting surfaces.  Vacuum
 * objects are places up- and down-stream from the thin plate.  These act as a
 * Particle sink.
 *
 * This simulation also demonstrates the new IntersectBoundaryObject class
 * that helps to get proper reflections/interactions correct near
 * intersections with multiple objects:  multiple reflections/interactions (within one
 * timestep) preserve the physical order.  
 */

/** Teleportation BoundaryObject.
 * Useful for implementing periodic boundary conditions.  */
template <class _Particle>
class TeleporterBoundaryObject : public virtual FlatBoundaryObject<_Particle> {
    typedef FlatSurface super0;
    typedef FlatBoundaryObject<_Particle> super1;
  public:
    TeleporterBoundaryObject(const Vector<double,3> & normal = V3(1.0,0.0,0.0),
                         const Vector<double,3> & point = 0.0,
                         const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                         const double & length_scaling = physical::unit::meter,
                         const Vector<double,3> & teleport_distance = 0.0 )
        : super0(), super1() {
        setSurface(normal,point,side_length,length_scaling,teleport_distance);
    }

    inline void setSurface(const Vector<double,3> & normal,
                           const Vector<double,3> & point,
                           const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                           const double & length_scaling = physical::unit::meter,
                           const Vector<double,3> & teleport_distance = 0.0 ) {
        /* a little redundant:  both super1 and super2 call
         * super0::setSurface.  Oh well. */
        const Vector<bool,3> allow_movement = true;
        super1::setSurface(normal, point, allow_movement, side_length, length_scaling);

        tdx = teleport_distance;
    }

    virtual ~TeleporterBoundaryObject()  {}

    /** Teleporting boundary.
     */
    virtual enum INTERACTION interact(_Particle & pold, _Particle & p) {
        double side_sign = 0;
        intersectTime(p,side_sign);

        if (side_sign <= 0) {
            /* we need to teleport the particle! */
            V3C(p.pos) += tdx;
            return CHANGED;
        } else return UNCHANGED;
    }

  private:
    Vector<double,3> tdx;
};

template <class _Particle>
class TeleWallBoundaryObject : public TeleporterBoundaryObject<_Particle>,
                               public WallBoundaryObject<_Particle> {
    typedef FlatSurface super0;
    typedef FlatBoundaryObject<_Particle> super1;
    typedef TeleporterBoundaryObject<_Particle> super2;
    typedef WallBoundaryObject<_Particle> super3;
  public:
    TeleWallBoundaryObject(const Vector<double,3> & normal = V3(1.0,0.0,0.0),
                           const Vector<double,3> & point = 0.0,
                           const Vector<double,3> & velocity = 0.0,
                           const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                           const double & length_scaling = physical::unit::meter,
                           const Vector<double,3> & teleport_distance = 0.0,
                           const double & temp = 300 * physical::unit::Kelvin,
                           const double & rough = -1,
                           const Vector<bool,3> & allow_movement = true,
                           const int & number_distrib_bins = 1000 )
        : super0(), super1(), super2(), super3() {
        setSurface(normal,point,velocity,side_length,length_scaling,teleport_distance,temp,rough,allow_movement,number_distrib_bins);
    }

    inline void setSurface(const Vector<double,3> & normal,
                           const Vector<double,3> & point,
                           const Vector<double,3> & velocity = 0.0,
                           const Vector<double,3> & side_length = V3(0.0,DBL_MAX,DBL_MAX),
                           const double & length_scaling = physical::unit::meter,
                           const Vector<double,3> & teleport_distance = 0.0,
                           const double & temp = 300 * physical::unit::Kelvin,
                           const double & rough = -1,
                           const Vector<bool,3> & allow_movement = true,
                           const int & number_distrib_bins = 1000 ) {
        /* a little redundant:  both super2 and super3 call
         * super0::setSurface and super1::setSurface.  Oh well. */
        super2::setSurface(normal, point, side_length, length_scaling, teleport_distance);
        super3::setSurface(normal, point, velocity, side_length, length_scaling, temp, rough, allow_movement, number_distrib_bins);
    }

    virtual ~TeleWallBoundaryObject()  {}

    /** Interact with wall and then teleport any that are on other side of
     * wall. 
     * FIXME:  This function should be tested to see if it actually works
     * correctly. */
    virtual enum INTERACTION interact(_Particle & pold, _Particle & p) {
        register bool r1 = super3::interact(pold,p) == CHANGED;/* wall */
        register bool r2 = super2::interact(pold,p) == CHANGED;/* teleporter */
        if (r1||r2) return CHANGED;
        else return UNCHANGED;
    }

};


}/* namespace. */

#endif // BOUNDARYOBJECT_H
