// -*- c++ -*-
// $Id: DSMCData.h,v 1.2 2005/01/19 04:31:21 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: DSMCData.h,v $
 * Revision 1.2  2005/01/19 04:31:21  olsonse
 * Added special trivial case of only one particle from which to obtain info such
 * as temperature.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */


/** \file
 * Particle type class definitions for holding various bits of data.
 * Copyright 2004 Spencer Olson.
 */


#ifndef DATAPARTICLE_H
#define DATAPARTICLE_H

#include <ostream>
#include <stdlib.h>
#include "common.h"
#include "Particle.h"
#include <olson-tools/memory.h>
#include <olson-tools/Vector.h>
#include "Octree.h"
#if (__sun__ != 1)
#  include <stdint.h>
#endif
#include <olson-tools/physical.h>

/** A particle class to hold data values in an octree.
 * */
template < class _Particle = Particle>
class DSMCData {
  public:
    /** Required position member for Octree\s. */
    double pos[3];

    /** we redef the new and delete operators so that we can cache these for
     * multiple tree building.
     */
    REDEF_NEW_DEL_H(DSMCData);

  public:
    /* Warning:  This enum MUST be in the same value order as the items in the
     * data struct!
     * */
    /** Enum of data supported by this data particle.
     * */
    enum DataType {
	NUMBER_DENSITY=0,
	NUMBER=1,
	MAX_SIGMA_RELSPEED=2,
	NUMBER_COLLISION_PAIRS_REMAINDER=3,
	NUMBER_COLLISIONS=4,
        AVG_V_X=5,
        AVG_V_Y=6,
        AVG_V_Z=7,
        AVG_V2_X=8,
        AVG_V2_Y=9,
        AVG_V2_Z=10,
        END_DATA=11
    };

    /** Data values of this data node. */
    union {
	/** A structure to access the data by name of data.
	*/
	struct {
            /** Average number density in the region represented by this data
             * node.
             * */
	    double numberDensity;

	    /** The total number of atoms represented by this data node.
	    */
	    double N;

	    /** The maximum relative speed of atoms represented by this data
	     * node.
	     */
	    double max_sigma_relspeed;

	    /** Collision pairs selection remainder.
	     */
	    double number_collision_pairs_remainder;

	    /** Number of collisions.
	     */
	    double collisions;

            /** Measured average velocity.  */
            double avgv[3];

            /** Measured average velocity^2.  */
            double avgv2[3];
	} name;
	/** An array to access the data by index.
	 * For clarity, use the enum values in enum DataType to index the data
	 * as an array.
	 * @see enum DataType
	 */
	double indx[];
    } data;

    /** Default constructor. */
    DSMCData();

    /** Copy constructor. */
    DSMCData(const DSMCData & d);

    DSMCData(const _Particle & q);

    /** A class for helping with building the data tree from
     * OctreeInterpolator::createDataOctree.
     */
    typedef struct InitInfo_s {
        /** a pointer to the list of Particle/s. */
        _Particle ** p;
    } InitInfo;

    /** Quick way to return effective radius.  Required by Octree. */
    inline double effective_radius() const {
        return data.name.N
               * 0.5
               * particle_info[particle_type].size
               * particle_info[particle_type].diameter;
    }

    /** Constructor primarily for use in building data tree with
     * OctreeInterpolator.
     */
    template <class OTreeType>
    inline DSMCData(const OTreeType & o, InitInfo & init) {
        zeroFields();
        /* set the position of this data point to the center of mass of the
         * particles in the octree node.  Not that it really matters.  This is
         * really only for debugging purposes.
         */
        o.copyRcm(pos);

        /* FIXME:  If ever multi-species runs are to be done, then we should
         * figure out how to handle this. */
        particle_type = init.p[0]->type;

        data.name.N = o.numberParticles();

        /* compute number density. */
        data.name.numberDensity = data.name.N / o.getVolume();

        if ( data.name.N == 1) {
            /* special case for node with only one super particle. */

            Vector<double,3> & v = V3C(init.p[o.low]->vel);
            Vector<double,3> & avgv = V3C(data.name.avgv);
            Vector<double,3> & avgv2 = V3C(data.name.avgv2);

            avgv = v;
            avgv2 = v; avgv2.compMult(v);

            data.name.max_sigma_relspeed = 0;

            return;
        }

        /* We do initializations, such as measure
         * density and so on.
         */

        /* loop through and sum up v and v^2 for sigma(v). */
        Vector<double,3> & vav = V3C(data.name.avgv);
        Vector<double,3> & v2av = V3C(data.name.avgv2);

        for (int i = o.low; i <= o.high; i++) {
            Vector<double,3> & v = V3C(init.p[i]->vel);
            vav += v;
            v2av += compMult(v,v);
        }

        /* finish avgv calc. */
        {
            register double OneOverN = 1.0 / data.name.N;
            v2av *= OneOverN;
            vav *= OneOverN;
        }
        Vector<double,3>  sigma2_v(v2av);
        sigma2_v -= compMult(vav,vav);

        /* come up with a maximum relative velocity for the Particle/s contained
         * in this node.  We'll try for a max speed of ~3\sigma_{v}.
         * */
        data.name.max_sigma_relspeed =
              particle_info[particle_type].cross_section
            * 3.0 * sqrt(sum(sigma2_v));
    }


    /** The radius of influence used in the octree interpolation algorithm.
     * */
    double radiusOfInfluence;

    /** The node in the data octree that is ``holding'' this data particle.
     * */
    octree::Octree<DSMCData> * octreeNode;

    /** Number of nearest neighbors.
     * Found by doing OctreeInterpolator::setRadiiOfInfluence(const int nNodes).
     * @see OctreeInterpolator::setRadiiOfInfluence(const int nNodes).
     * */
    int nearestNeighbors;

    /** Helps in initialization of the class.
     * Just zeros all fields special to the DSMCData.  (This does not
     * touch the Particle fields.
     * */
    inline void zeroFields() {
        radiusOfInfluence = 0;
        octreeNode = NULL;
        memset(&data, 0, sizeof (data));
        nearestNeighbors = 0;
        particle_type = 0;
        next = NULL;
    }

    /** An index into the particle_info database.
     * FIXME:  is this the best place for this?  For the general case where
     * there are multiple specicies in the simulation, should this be an
     * array?
     */
    int particle_type;

    /** precscribed to be useable in OctreeMonitor. */
    inline const int & particleType() const {return particle_type;}

    /** Pointer to next Particle in a linked list of Particle\s.
     */
    class DSMCData * next;

    inline Vector<double,3> temperature() const {
        Vector<double,3> & vav = V3C(data.name.avgv);
        Vector<double,3> & v2av = V3C(data.name.avgv2);
        Vector<double,3> T;
        T = v2av; T -= compMult(vav,vav);
        T *= particle_info[particle_type].mass / physical::constant::K_B;

        return T;
    }

    inline double avgTemperature() const {
        return mean(temperature());
    }

    /** startMeasure is prescribed for use in OctreeMonitor. */
    template <class T>
    inline void startMeasure(const T & o, const int & ptype) {
        zeroFields();

        /* set the position of this data point to the center of mass of the
         * particles in the octree node.  Not that it really matters.  This is
         * really only for debugging purposes.
         */
        o.copyRcm(pos);

        /* FIXME:  If ever multi-species runs are to be done, then we should
         * figure out how to handle this. */
        particle_type = ptype;
    }

    /** addMeasure is prescribed for use in OctreeMonitor. */
    inline void addMeasure(const _Particle & p) {
        data.name.N++;
        Vector<double,3> & v = V3C(p.vel);
        Vector<double,3> & vav = V3C(data.name.avgv);
        Vector<double,3> & v2av = V3C(data.name.avgv2);

        vav += v;
        v2av += compMult(v,v);
    }

    /** finishMeasure is prescribed for use in OctreeMonitor. */
    template <class T>
    inline void finishMeasure(const T & o) {
        /* finish avgv calc. */
        if (!data.name.N) return;

        double OneOverN = 1.0 / data.name.N;

        Vector<double,3> & vav = V3C(data.name.avgv);
        Vector<double,3> & v2av = V3C(data.name.avgv2);
        Vector<double,3> sigma_v;

        vav *= OneOverN;
        v2av *= OneOverN;

        sigma_v = v2av; sigma_v -= compMult(vav,vav);

        /* come up with a maximum relative velocity for the Particle/s contained
         * in this node.  
         * */
        data.name.max_sigma_relspeed =
              particle_info[particle_type].cross_section
            * 3.0 *sqrt(sum(sigma_v));

        /* compute number density. */
        data.name.numberDensity = data.name.N / o.getVolume();

    }/* finishMeasure */

};

/** Overloaded stream output for DSMCData\s.
 * */
template < class _Particle>
std::ostream & operator<<(std::ostream & output,const DSMCData<_Particle> & q);

/** Overloaded stream input for DSMCData\s for use with readData and
 * readDataBlocks.
 * @see readData and readDataBlocks.
 * */
template < class _Particle>
std::istream & operator>>(std::istream & input, DSMCData<_Particle> * & d);

#include "DSMCData.C"

#endif // DATAPARTICLE_H
