// -*- c++ -*-
// $Id: FieldData.h,v 1.3 2005/06/07 10:00:07 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: FieldData.h,v $
 * Revision 1.3  2005/06/07 10:00:07  olsonse
 * Many changes over the course of time
 *
 * Revision 1.2  2005/01/19 04:32:07  olsonse
 * Added compile flag to enable compile on SunOS.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 */


/** \file
 * Particle type class definitions for holding field data.
 * Copyright 2004 Spencer Olson.
 */


#ifndef FIELDDATA_H
#define FIELDDATA_H

#include <ostream>
#include <istream>
#include <stdlib.h>
#include "common.h"
#include <olson-tools/memory.h>
#include "Octree.h"
#if (__sun__ != 1)
#  include <stdint.h>
#endif
#include <vector>
#include <olson-tools/Vector.h>


/** A particle class to hold data values in an octree.
 * */
class FieldData {
  public:
    /** Required position member for Octree\s. */
    double pos[3];

    /** we redef the new and delete operators so that we can cache these for
     * multiple tree building.
     */
    REDEF_NEW_DEL_H(FieldData);

  public:
    /* Warning:  This enum MUST be in the same value order as the items in the
     * data struct!
     * */
    /** Enum of data supported by this data particle.
     * */
    enum DataType {
	FIELD_X=0,
	FIELD_Y=1,
	FIELD_Z=2,
	FIELD_MAG=3,
	FIELD_GRAD=4
    };

    /** Data values of this data node. */
    union {
	/** A structure to access the data by name of data.
	*/
	struct {
	    /** Field components.
	     * To be clear and consistent, index this array by using the X,Y,Z macros
	     * defined in common.h.
	     */
	    double field[3];

            /** Field Magnitude. */
            double magnitude;

            /** Field Gradient. */
            double gradient;
	} name;
	/** An array to access the data by index.
	 * For clarity, use the enum values in enum DataType to index the data
	 * as an array.
	 * @see enum DataType
	 */
	double indx[];
    } data;

    /** Default constructor. */
    FieldData();

    /** Copy constructor. */
    FieldData(const FieldData & d);

    /** Init constructor. */
    inline FieldData(const Vector<double,3> & x, const Vector<double,3> & F) {
        zeroFields();
        memcpy(pos, x.val, 3*sizeof(double));
        memcpy(data.name.field, F.val, 3*sizeof(double));
        data.name.magnitude = F.abs();
        data.name.gradient = 0;
    }

    /** Constructor that doesn't seem too relevant yet.
     * This constructor is implemented simply because the OctreeInterpolator
     * requires it.
     */
    FieldData(const octree::Octree<FieldData> & o, FieldData * a[]);

    /** The radius of influence used in the octree interpolation algorithm.
     * */
    double radiusOfInfluence;

    /** The node in the data octree that is ``holding'' this data particle.
     * */
    octree::Octree<FieldData> * octreeNode;

    /** Number of nearest neighbors.
     * Found by doing OctreeInterpolator::setRadiiOfInfluence(const int nNodes).
     * @see OctreeInterpolator::setRadiiOfInfluence(const int nNodes).
     * */
    int nearestNeighbors;

    /** Helps in initialization of the class.
     * Just zeros all fields special to the FieldData.  (This does not
     * touch the Particle fields.
     * */
    inline void zeroFields() {
        radiusOfInfluence = 0;
        octreeNode = NULL;
        memset(&data, 0, sizeof (data));
        nearestNeighbors = 0;
        next = NULL;
    }

    /** Pointer to next Particle in a linked list of Particle\s.
     */
    class FieldData * next;

};

/** Overloaded stream output for FieldData\s.
 * */
std::ostream & operator<<(std::ostream & output,const FieldData & q);

/** stuff for reading FieldData from streams. */
class  FieldValue {
  public:
    double val[8];
};

/** Read FieldData from an input stream.
 * @returns
 *     A linked list of FieldData.
 */
FieldData * readFieldData(std::istream & input, std::vector<FieldValue> * V = NULL);

#endif // FIELDDATA_H
