// -*- c++ -*-
// $Id$
/*
 * Copyright 2005 Spencer Olson
 *
 * $Log$
 *
 */

#ifndef FLATSURFACE_H
#define FLATSURFACE_H

#include <cfloat>
#include <olson-tools/Vector.h>
#include <olson-tools/m_eps.h>
#include "common.h"

class FlatSurface {
  public:
    /** FlatSurface constructor.
     * @param normal
     *     Normal to the surface of the wall.
     * @param point
     *     A point lying on the wall somewhere.
     * @see setSurface.
     */
    inline FlatSurface(const Vector<double,3> & normal = V3(1.0,0.0,0.0),
                       const Vector<double,3> & point = V3(0.0,0.0,0.0),
                       const Vector<bool,3> & allow_movement = true ) {
        setSurface(normal,point,allow_movement);
    }

    /** Bulldozing. */
    inline ~FlatSurface() {}

    /** Time until intersection.
     * This function never tests for the case when Particle/s are moving
     * towards the wall on the wrong side of the wall; this of course w/could be
     * an error as it is not necessarily physical--unless we want this code to
     * implement a one-way Wall.
     *
     * And now do an active rotation for reflecting the particles
     * position.  We first must determine the point of intersection of the
     * particle with the wall.
     * for:
     *  -   t :  time of intersection
     *  -   r :  position of particle
     *  -   v :  velocity of particle
     *  -   p :  a point on the reflecting wall
     *  -   n :  normal to reflecting wall
     *  .
     *      t = \frac{ n \cdot (p - r) } { n \cdot v }
     *
     *
     * @returns The time (in seconds) Particle p intersects with the
     * wall---+DBL_MAX is returned for parallel trajectories.
     * t < 0 indicates an intersection in the past. 
     */
    template <class _Particle>
    inline double intersectTime(const _Particle & part) const {
        double side_sign;
        return intersectTime(part,side_sign);
    }

    /** Time until intersection including side of surface.
     * @param part (input)
     *     The particle that we are to test for intersection.
     *
     * @param side_sign (output)
     *     On return indicates the side (relative to the normal specified) of
     *     the surface the particle is currently on.  If the particle is on
     *     the side facing the normal, side_sign is set to 1.0 and to -1.0 if
     *     on the opposite side.  If the particle is within M_EPS of the
     *     surface, t is returned as zero and side_sign is set to 0.
     *
     * @see intersectTime(const _Particle & part).
     */
    template <class _Particle>
    inline double intersectTime(const _Particle &  part, double & side_sign) const {
        Vector<double,3> & v = *((Vector<double,3>*)part.vel);
        Vector<double,3> & r = *((Vector<double,3>*)part.pos);
        register double n_dot_r = (n * r);

        /** for testing whether the particle is ON the surface:
         * epsilon == (10 * M_EPS) * 0.5*( (n * r) + (n * p) )
         * The extra factor of 10 on M_EPS is for safety to keep us away from
         * precision error--like it will help any way...
         */
        register double epsilon = 5.0 * M_EPS * fabs(n_dot_r + n_dot_p);

        /* FIXME:  Perhaps we should just add a thickness to the flat surfaces
         * so that we can ensure that particles cannot possibly be confused as
         * to which side of the flat surface they belong.  */
        /* we have to let fence sitters be.  Take care of deciding which side
         * the particle is really on using another call to intersectTime with
         * past/current particle.  */
        register double n_dot_r_minus_p = (n_dot_r - n_dot_p);
        if (fabs(n_dot_r_minus_p) <= epsilon) {
            n_dot_r_minus_p = 0.0;
            side_sign = 0.0;
        } else if (n_dot_r_minus_p >= 0.0) side_sign = +1.0;
        else side_sign = -1.0;

        /* FIXME:  how small to let ndotv go? */
        register double ndotv = n * v;
        if ( ndotv == 0 ) return DBL_MAX;
        else return -n_dot_r_minus_p / ndotv;
    }

    template <class _Particle, class _Container>
    inline double intersectTime(const _Particle & particle_old,
                                const _Particle & particle,
                                const _Container & volb ) const {

        double side_sign = 0.0;
        Vector<double,3> iloc;
        return intersectTime(particle_old, particle, volb, iloc, side_sign );
    }

    /** Runs the full intersection test to see if an intersection occurred
     * between a past and present instance of a particle.  Returns the
     * intersection time (negative) relative to the present instance or
     * DBL_MAX if no intersection occurred between the past and present.  
     *
     * Note that the convention here is that intersections occur if t <=0.
     * Thus, if a past instance resides on the surface such that its time
     * distance to the surface is zero, the intersection is considered to have
     * already occurred in a previous step and will thus be ignored in the
     * current test. 
     *
     * @param particle_old (input)
     *     The particle before the last move.
     *
     * @param particle (input)
     *     The current particle.
     *
     * @param volb (input)
     *     The bounding volume of this surface that is to be tested.  
     *
     * @param iloc (output)
     *     Location of the intersection IF the return value is finite
     *     (otherwise, iloc is undefined).
     *
     * @param side_sign (output)
     *     On return indicates the side (relative to the normal specified) of
     *     the surface the particle was before traversal through the surface
     *     (if that did indeed happen).  If the particle was on the side facing
     *     the normal, side_sign is set to 1.0 and to -1.0 if on the opposite
     *     side.  This value is undefined if the function returns t=DBL_MAX.
     *
     * @returns
     *     The time since the intersection relative to the current particle
     *     position.  If the intersection is determined to not have happened
     *     between the old and current particle instances, then t is returned
     *     as DBL_MAX.   
     *
     * @see intersectTime(const _Particle & part, double & side_sign).
     * @see Wall::reflect.
     */
    template <class _Particle, class _Container>
    inline double intersectTime(const _Particle & particle_old,
                                const _Particle & particle,
                                const _Container & volb,
                                Vector<double,3> & iloc,
                                double & side_sign ) const {

        /* We first must determine the point of intersection of the
         * particle with the wall.
         * for:
         *      t :  time of intersection
         *      r :  position of particle
         *      v :  velocity of particle
         *      p :  a point on the reflecting wall
         *      n :  normal to reflecting wall
         *
         *      t = \frac{ n \cdot (p - r) } { N \cdot v }
         */
        /* get the intersect time of the past particle. */
        double t = intersectTime(particle_old,side_sign);
        if ( !finite(t) || t <= 0 ) {
            /* parallel movement or particle prior to last movement already
             * past the surface. 
             * Note that intersectTime above already clamps the time to zero
             * when within M_EPS of zero. */
            return DBL_MAX;
        }

        /* now get intersect time of the current particle. */
        t = intersectTime(particle);

        /* protected velocity:  v . allowed_movement
         * used to only move particle in allowed directions.
         */

        if (!volb.contains(intersectLocation(particle,t,iloc).val)) {
            /* The intersection point is actually not on the defined surface. */
            return DBL_MAX;
        }

        /* passed all of the tests.  Note that this can be both +/- for
         * future/past. */
        return t;
    }

    /** Intersection location.
     * 
     * @param part
     *     The particle to test (this _Particle requires .vel and .pos
     *     3-vectors).
     *
     * @param t
     *     return value of intersectTime.
     *
     * @param loc
     *     The return value is just pos += vel*t (only meaningful for times
     *     returned by intersectTime).
     *
     * @see intersectTime
     */
    template <class _Particle>
    inline const Vector<double,3> & intersectLocation(const _Particle & part,
                                                      const double & t,
                                                      Vector<double,3> & loc) const {
        loc  = V3C(part.vel);
        loc.compMult(allowed_movement);
        /* trying to ensure that we don't let precision problems confuse which
         * side we are on.  This essentially makes all surfaces defined to a
         * precision of (10*M_EPS) of the average particle transit. */
        loc *= t * (1.0 - copysign(10.0*M_EPS,t));
        loc += V3C(part.pos);
        return loc;
    }

    /** Initialize the class for work.
     * This must be done AFTER the parameters are chosen, and
     * BEFORE any atoms can be inserted into the module.
     * Note that a velocity component of the wall normal to the wall may be a
     * bit problematic.
     * @param normal
     *     Normal to the surface of the wall.
     * @param point
     *     A point lying on the wall somewhere.
     * */
    inline void setSurface(const Vector<double,3> & normal,
                           const Vector<double,3> & point,
                           const Vector<bool,3> & allow_movement = true) {
        /* let's make sure that surface_direction vector is normalized */
        n = normal/sqrt(normal*normal);
        p = point;

        /* inner product n*p is used for intersection test. */
        n_dot_p = (n * p);

        allowed_movement = allow_movement;
    }

    /** Access to a point lying on the surface. */
    inline const Vector<double,3> & point() const { return p; }

    /** Access to normal to the surface. */
    inline const Vector<double,3> & normal() const { return n; }

    /** Access to allowed movement vector. */
    inline const Vector<bool,3> & movable() const { return allowed_movement; }

  protected:
    Vector<double,3> n;
    Vector<double,3> p;

    Vector<bool,3> allowed_movement;

    /** Inner product of n_dot_p used for intersection test. */
    double n_dot_p;
};

#endif // FLATSURFACE_H
