// -*- c++ -*-
// $Id: OctreeContainer.h,v 1.2 2005/06/07 10:00:11 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: OctreeContainer.h,v $
 * Revision 1.2  2005/06/07 10:00:11  olsonse
 * Many changes over the course of time
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 */

#ifndef OCTREECONTAINTER_H
#define OCTREECONTAINTER_H

#include "Octree.h"
#include <stdexcept>
#include <vector>
#include "BoundaryObject.h"
#include <olson-tools/ompexcept.h>
#include <olson-tools/logger.h>

namespace octree {

/** Auxillary class for removing Particle/s correctly with OctreeContainer/s.
*/
class RemovalQueueNode {
    REDEF_NEW_DEL_H(RemovalQueueNode);
  public:
    RemovalQueueNode(const int & i) : idx(i), next(NULL) {}
    int idx;
    RemovalQueueNode * next;
};

/* NOTE:
 * I mixed the new/delete calls with realloc/free calls in this class.
 * I use realloc because it makes it more convenient to reallocate space for
 * the Particle list as it grows.
 */

/** An implementaion of a Particle and Octree container.
*/
DefaultOctreeTemplate
class OctreeContainer {
    typedef OctreeContainer<OctreeTemplateArgs> mytype;
  public:
    typedef Octree<OctreeTemplateArgs> OctreeType;

    /** The space increment amount for Particle\s array.
     * When additional space is needed to store pointers to the Particle\s
     * held by this container, the Particle\s array is realloc()'d to add
     * this amount of extra space.
     */
    static int particleSpaceIncrement;

    /** OctreeContainer style:  either a temporary container or the owner of its
     * particles.
     */
    enum STYLE {
        /** OWNER style specifies that the OctreeContainer can/will delete particles
         * when necessary, such as when the OctreeContainer::~OctreeContainer
         * method is called.
         */
        OWNER,
        /** TEMPORARY style specifies that the OctreeContainer is only a
         * temporary container and mustn't change/delete its particles.
         */
        TEMPORARY
    };

    /** Default constructor.
     * @param o
     *     Specify either OWNER or TEMPORARY OctreeContainer STYLE.
     */
    OctreeContainer(enum STYLE o = OWNER);

    /** Destructor. */
    ~OctreeContainer();

    /** Add Particles to the container.
     * @param input
     *     A linked list of particles.
     * @returns
     *     The number of particles added to the container.
     * @throws
     *     std::bad_alloc if a call to realloc() fails.
     */
    int addParticles(_Particle * input) THROWS(std::bad_alloc);

    /** Get a pointer to the Particle pointer array.
     * @return A pointer to the array of pointers to the Particle\s held by the
     * OctreeContainer.
     */
    inline _Particle ** getParticles() const { return particles; }

    /** Get the total number of Particle\s held by the OctreeContainer.
     * @return Number of Particle\s held by the OctreeContainer.
     */
    inline const int & numberParticles() const { return nParticles; }

    /** Remove the pointer to a Particle from the Particle/s table.
     * Note that this function moves moves the last particle from the total
     * particle list to the position of the removed particle pointer.  The
     * number of particles in the OctreeContainer is then decremented.
     * In other words, the underlying Octree is no longer valid after use of
     * this function.
     */
    inline void removePtr(const int & i) {
        /* setting it to NULL takes care of the case when we delete the
         * last one.
         * NOTE:  mov %eax,0 instruction is better than j* %eax
         *    (that is why there is no:
         *       if(last-element) particles[--nParticles] = NULL;
         *       else {
         *           particles[i]  = particles[--nParticles];
         *           particles[nParticles] = NULL;
         *       }
         *    )
         */
        particles[i] = NULL;
        particles[i] = particles[--nParticles];

        /* set the previously last particle pointer to NULL.  This is for
         * helping the debug process.
         */
        particles[nParticles] = NULL;
    }

    /** Queue the removal of the pointer to a given Particle. */
    inline void queuePtrRemoval(const int & i) {
        RemovalQueueNode * n = new RemovalQueueNode(i);
        n->next = removal_queue;
        removal_queue = n;

        /* the  pointer must now be NULL */
        particles[i] = NULL;
    }

    /** Remove the pointer to a Particle from the Particle/s table.
     * WARNING:  removePtr(int) and queuePtrRemoval/flushRemovalQueue are NOT
     * compatible and should not happen together (removePtr should not be
     * called between queuePtrRemoval and flushRemovalQueue.
     *
     * Again, the Octree is invalid after this
     * operation.
     *
     * @see removePtr(const int & i).
     */
    inline void flushRemovalQueue() {
        int last = nParticles-1;
        while (removal_queue) {
            /* first find the last good pointer.
             * This scheme relies on queuePtrRemoval to set
             * particles[i] = NULL.
             */
            while (last >=0 && particles[last] == NULL) last--;

            /* we only fix a pointer if it is < last which.  Note that
             * last >= nParticles(the final value).
             */
            if (removal_queue->idx < last) {
                particles[removal_queue->idx] = particles[last];
                particles[last] = NULL;
            }
            nParticles--;


            register RemovalQueueNode * next = removal_queue->next;
            delete removal_queue;
            removal_queue = next;
        }
    }

    /** Add a boundary object.
     * Note that the order with which boundaries are added is important.  The
     * 'add order' determines also the order by which the particles are
     * interacted with the boundaries.  Therefore, if an earlier boundary
     * determines that the particle needs to be removed from the system, it
     * will not interact with later boundaries.
     */
    inline void addBoundaryObject(BoundaryObject<_Particle> * boPtr) {
        boundary_objects.push_back(boPtr);
    }

    /** Remove a boundary object (ALL instances/pointers thereto).
     */
    inline void removeBoundaryObject(BoundaryObject<_Particle> * boPtr) {
        boundary_objects.push_back(boPtr);
        for (typename std::vector<BoundaryObject<_Particle> *>::iterator i = boundary_objects.begin();
             i < boundary_objects.end(); i++) {
            if ((*i) == boPtr) {
                i = boundary_objects.erase(i);
                i--;
            }
        }/* for */
    }

    /** Apply boundary conditions from BoundaryObject/s.
     *  This function calls a private recursive routine that finds the
     *  smallest Octree branch that interacts with a certain boundary.
     *  TODO:  make sure that this routine only recurses as far as it needs to
     *  (as mentioned here in the description. )
     *
     * @return Number of particles removed from container.
     */
    inline int interactParticlesWithBoundaryObjects(const double dx[]) {
        int retval = interactParticlesWithBoundaryObjectsR(*getOctree(), dx);

        /* make sure the removal queue was emptied first. */
        flushRemovalQueue();
        return retval;
    }

  private:

    int interactParticlesWithBoundaryObjectsR(OctreeClass & o, const double dx[]) {
        if (o.thisHasChildren) {

            /* not a leaf node, so continue recursing to children */
            int nremoved = 0;
            for (int i = 0; i < 8; i++) {
                if (o.child[i]) {
                    nremoved += interactParticlesWithBoundaryObjectsR(*o.child[i], dx);
                }
            }
            return nremoved;

        } else {
            /* have the container apply boundary conditions. */
            return interactParticlesWithBoundaryObjects(o, dx);
        }/* else */
    }

  public:

    /** This function prepares a vector of valid BoundaryObjects for a
     * specific octree node.  The caller is responsible for doing things in
     * the correct order and using the returned vector appropriately. 
     *
     * @param out
     *     On return, this will contain a list BoundaryObjects valid for this
     *     Octree node (note that out.clear() will be called).
     *
     * @param o
     *     The Octree node of interest.
     *
     * @param dx
     *     The amount by which to scale BoundaryObject bounding volumes (this
     *     allows for some amount of jitter for Particle location, as is the
     *     case when particles are moving around).
     */
    inline void minimizeBoundaryObjects( std::vector<BoundaryObject<_Particle> *> & out,
                                         OctreeClass & o,
                                         const double dx[]) {
        out.clear();

        int mboi = boundary_objects.size();
        for ( int i = 0; i < mboi; i++) {
            BoundaryObject<_Particle> & bo = *boundary_objects[i];

            /* first allow boundaries to scale their volume. */
            bo.scaleVolume(dx);

            /* now do an overlap test with this boundary and this Octree node. */
            if (bo.volb_scaled.overlaps(o.getVolB())) {
                out.push_back(&bo);
            }
        }
    }

    /** Apply boundary conditions from BoundaryObject/s to a single Particle.
     *  Note that particles that are removed from this OctreeContainer as a
     *  result of interacting with a particular BoundaryObject can be deleted
     *  and may not be contained in memory after this function completes.  One
     *  MUST wait until flushRemovalQueue() is called before particles[] list
     *  might be valid.
     *
     * The boundary_objects list is a result of calling
     * minimizeBoundaryObjects().
     *
     * @returns Whether the particle was removed from the system.
     *
     * @see minimizeBoundaryObjects().
     */
    inline bool interactParticleWithBoundaryObjects(
                        _Particle & pold,
                        const int & pidx,
                        std::vector<BoundaryObject<_Particle> *> & bolist) {
        _Particle & p = *particles[pidx];

        int mboi = bolist.size();
        for ( int i = 0; i < mboi; i++) {
            BoundaryObject<_Particle> & bo = *bolist[i];

            /* FIXME:  (actually a TESTME) it seems that we need to allow all
             * particles to be tested so that the boundary object can
             * determine whether a the corner of the object might be clipped.
             *
             * Some time, I need to re-enable to corner code.  Perhaps there
             * should be a place (hook) allowed here to reorder the
             * interactions with various boundaries.  There could be a routine
             * to do this for boundaries that are coupled.
             * ---
             * For right now, I will just let boundaries interact in the order
             * they are added to the container.
             */
            //if (!bo.volb_scaled.contains(p.pos)) { continue;/* not contained. */ }

            /* the particle is contained, so allow it to interact
             * with the boundary. */
            switch (bo.interact(pold,p)) {
                default : case UNCHANGED: break;

                case CHANGED:
                    /* interact() was supposed to change pold so that it
                     * reflects the new 'old' particle.  */
                    break;

                case REMOVED:
                    /* the particle needs to be removed from the system.
                     * First properly remove it from the particle list.*/
                    queuePtrRemoval(pidx);

                    if (container_style == OWNER) {
                        /* only the owner can actually change _Particle
                         * pointers. */

                        if (bo.keepRemoved) {
                            /* we are to place the _Particle in the buffer of the
                             * boundary. */
                            p.next = bo.removed;
                            bo.removed = &p;
                        } else {
                            delete &p;
                        }

                    }/* if OWNER */

                    /* the working assumption is that we don't have to let more
                     * than one boundary remove the same particle. */
                    return true;
            }/* switch */
        }/* for boundaries */

        return false;
    }/* interactParticlesWithBoundaryObjects */

    /** Apply boundary conditions from BoundaryObject/s.
     *  Note that particles that are removed from this OctreeContainer as a
     *  result of interacting with a particular BoundaryObject can be deleted
     *  and may not be contained in memory after this function completes.  One
     *  MUST wait until flushRemovalQueue() is called before particles[] list
     *  might be valid.
     *
     * @return Number of particles removed from container.
     *
     *  @see interactParticlesWithBoundaryObjects(const double dx[]) for a
     *       recursive call for all nodes. 
     */
    inline int interactParticlesWithBoundaryObjects(OctreeClass & o, const double dx[]) {
        std::vector<BoundaryObject<_Particle> *> bo_overlap;

        minimizeBoundaryObjects(bo_overlap, o, dx);

        int nremoved = 0;
        /* loop through each particle. */
        for (int j = o.low; j <= o.high; j++) {
            /* first do a containment test for this boundary. */
            _Particle pold = *particles[j];

            nremoved += interactParticleWithBoundaryObjects(pold, j, bo_overlap);
        }/* for _Particles */

        return nremoved;
    }/* interactParticlesWithBoundaryObjects */

    /** Allocates and builds the octree for sorting Particle\s.
     * Note that any previously built octree's should first be free'd with
     * freeOctree() and the particle pointer removal queue is flushed with
     * flushRemovalQueue().
     *
     * @param min_part
     *    The minimum number of particles per Octree node (this is a
     *    non-strict minimum).
     * 
     * @see Octree::buildOctree.
     */
    inline void buildOctree(const LeafModel & leafModel = LeafModel()) {
	octree = new OctreeClass(0, nParticles-1,0);
	octree->buildOctree(particles, leafModel);
    }

    /** Frees the particles held by the container.
     * If this OctreeContainer is NOT an OctreeContainer::OWNER, then this
     * function will only free the particle table.
     */
    inline void freeParticles() {
        if (particles) {
            /* delete all particles in the list */
            if (container_style == OWNER) {
                for (int i = 0; i < nParticles; delete particles[i++]);
            }

            nParticles = 0;

            /* and delete the particle table.
             * REMEMBER:  particles is allocated via realloc() not new[].
             */
            free(particles);
            particles = NULL;

            /* reset the space allocated register. */
            particleSpaceAllocated = 0;
        }/* if */
    }

    /** Frees memory used by the Particle octree.
    */
    inline void freeOctree() {
	if (octree) {
	    delete octree;
	    octree = NULL;
	}
    }

    /** Get a pointer to the Particle octree.
     * @return pointer to the internally stored Particle octree.
     */
    inline OctreeClass * getOctree() const { return octree;}

    /** Copy the center of mass position to an array.*/
    inline void copyRcm(double  p[]) const {
        octree->copyRcm(p);
    }

    /** Re-measure center of mass position and bounding volume for
     * the root node only. 
     * NOTE:  this saves the root-node measurement with the root node, but
     * since the leaves are not changed, the information is no-longer
     * coherent.  This also updates the low,high range for the root node,
     * further disconnecting the root from the leaves.  Because of this
     * disconnect, we're going to just free the Octree. 
     */
    inline void refreshCurrentRcmAndVol() {
        freeOctree();
	octree = new OctreeClass(0, nParticles-1,0);
        octree->findRcmAndVol(particles);
    }


    /** Measure center of mass position and bounding volume for
     * the root node only. 
     * NOTE:  This DOES NOT save the newly made measurements to anything. 
     */
    inline void measureCurrentRcmAndVol(vol_bounds & volb) const {
	OctreeClass * tmpoctree = new OctreeClass(0, nParticles-1,0);
        tmpoctree->findRcmAndVol(particles);
        volb = tmpoctree->getVolB();
    }

    /** Return the top Octree node's bounded volume. */
    inline const vol_bounds & getVolB() const { return octree->getVolB(); }

    /** Resets the container to hold a specific number of particles.
     * NOTE:  clears Particle\s and Octree.
     *
     * ALSO NOTE:  This function DOES NOT initialize the particle pointer table.
     *     The calling function is expected to fill in this table before it
     *     might be accessed resulting in SEGV.
     */
    inline void setNumberParticles(const int & n) THROWS(std::bad_alloc) {
        /* because we are resizing so much, let's just assume we are supposed
         * to start from scratch.  So let's free everything. */
        flushRemovalQueue();
        freeParticles();
        freeOctree();

        nParticles = n;

        /* allocate space for the particles */
        particleSpaceAllocated = nParticles;
        particles = (_Particle **)realloc(particles, particleSpaceAllocated * sizeof(void*));
        if (particles == NULL) THROW(std::bad_alloc,);

        /* allocate tree root.
         * We'll just assume that the caller is going to fill in the actual
         * particles later on so that this new Octree root node won't be
         * invalid for long.
         */
        octree = new OctreeClass(0, nParticles-1,0);
    }

  public:

    /** Plot the tree and particles into the same gnuplot output.
    */
    std::ostream & plotParticlesAndTree(std::ostream & output,
                                  const std::string & title,
                                  const std::string & tree_plotopt,
                                  const std::string & part_plotopt) const;

    /** Write particles into output stream delimited by newlines.
     * @see operator<<(std::ostream & output,const Particle & q).
    */
    inline std::ostream & printParticles(std::ostream & output) const {
        for (int i = 0; i<nParticles; i++) { 
            output << *(particles[i]) << '\n';
        }
        return output;
    }

    inline const std::vector<BoundaryObject<_Particle> *> & boundaryObjects() const {
        return boundary_objects;
    }

    /** A simple function to help debugging null pointers in the particle
     * list.
     * This function is a nice utility for using inside the debugger.
     *
     * @return The number of null-pointers in the particle list (this should
     * be zero).
     */
    void check_particle_list() const {
        check_particle_array(particles,0,nParticles);
    }

  private:
    _Particle ** particles;
    int nParticles;
    OctreeClass * octree;
    int particleSpaceAllocated;

    enum STYLE container_style;

    std::vector<BoundaryObject<_Particle> *> boundary_objects;

    RemovalQueueNode * removal_queue;

};

}/* namespace octree. */

#include "OctreeContainer.C"

#endif // OCTREECONTAINTER_H

