// -*- c++ -*-
// $Id: OctreeInterpolator.C,v 1.4 2005/06/07 10:00:11 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: OctreeInterpolator.C,v $
 * Revision 1.4  2005/06/07 10:00:11  olsonse
 * Many changes over the course of time
 *
 * Revision 1.3  2005/02/04 19:55:45  olsonse
 * Added few lines (semi-experimental) to readjust the radius of influence for
 * each data node after finding the nearest neighbors.  It is scaled as
 * R = (nearest nodes wanted / nearest nodes found)^(1/3) * (maximum distance to nearest neighbors)
 *
 * Revision 1.2  2005/01/19 11:39:16  olsonse
 * Changed interpolator algorithm to be able to interpolate several data values
 * at the same time.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */

#include "OctreeInterpolator.h"
#include <math.h>
#include <olson-tools/logger.h>
#include <olson-tools/ompexcept.h>
#include <stdexcept>

namespace octree {

template <class _DataParticle, class _Particle>
OctreeInterpolator<_DataParticle,_Particle>::OctreeInterpolator():
        OctreeContainer<_DataParticle,InterpolatorFields>(),
        W(NULL),
        dr_h_recip(0) {
        initKernel();
};

template <class _DataParticle, class _Particle>
OctreeInterpolator<_DataParticle,_Particle>::~OctreeInterpolator() {
    if (W) {
        delete[] W;
        W = NULL;
    }
}

template <class _DataParticle, class _Particle>
template <class PTreeType, class T>
void OctreeInterpolator<_DataParticle,_Particle>::createDataOctree(PTreeType * ptree, T & auxillary) THROWS(std::runtime_error) {
    /* this can only be successfully run if the particle octree already
     * exists
     */
    if (ptree->getVolRadius() <= 0)
        THROW(std::runtime_error,FILE_LINE "findRcmAndVol(...) not run on octree yet.");

    /* reset the number of particles allowed in the container.
     * NOTE:  this function clears the Particle list and Octree.
     */
    setNumberParticles(ptree->leaves);

    /* now recursively build the tree and copy over the info */
    copyDataFromParticleOctree(*super::getOctree(), *ptree, auxillary);
}

template <class _DataParticle, class _Particle>
template <class PTreeType, class T>
void OctreeInterpolator<_DataParticle,_Particle>::copyDataFromParticleOctree(
                            Octree<_DataParticle,InterpolatorFields> & dtree,
                            const PTreeType & ptree,
                            T & auxillary) {
    /* copy over appropriate tree stuff from the particle octree node */
    dtree.leaves = ptree.leaves;
    dtree.copyRcmAndVol (ptree);

    /* copy data into the data particle ONLY if this is a leaf */
    if (dtree.leaves == 1) {
	/* allocate data particle
	 * Note that this contsructor takes care of initializing the data from
	 * ptree.
	 */
        _DataParticle * d = new _DataParticle(ptree, auxillary);
        d->octreeNode = (Octree<_DataParticle>*)&dtree;
	super::getParticles()[dtree.low] = d;
	dtree.thisHasChildren = false;

    } else {
	dtree.thisHasChildren = true;
	int lo = dtree.low;

	for (int i = 0; i < 8; i++) {
	    if (ptree.child[i]) {
		int lo2 = (lo+ptree.child[i]->leaves);
		dtree.child[i] = new Octree<_DataParticle,InterpolatorFields>(lo, lo2-1, ptree.child[i]->depth);
		lo = lo2;
		dtree.child[i]->parent = &dtree;
		copyDataFromParticleOctree(*dtree.child[i], *ptree.child[i], auxillary);
	    }/* if */
	}/* for */
    }/* else */
}

template <class _DataParticle, class _Particle>
bool OctreeInterpolator<_DataParticle,_Particle>::createDataOctree(_DataParticle * data) {
    /* make sure that we are not copying over an old tree */
    super::flushRemovalQueue();
    super::freeParticles();
    super::freeOctree();

    super::addParticles(data);

    super::buildOctree(1);

    /* now because we are building the tree directly from the data, we need to
     * walk the tree and link all data nodes to their Octree leaf.
     */
    relinkDataParticlesToOctree(*super::getOctree());

    return true;
}

template <class _DataParticle, class _Particle>
void OctreeInterpolator<_DataParticle,_Particle>::relinkDataParticlesToOctree(Octree<_DataParticle,InterpolatorFields> & dtree) {
    if (dtree.leaves == 1) {
        /* This is a leaf, so copy the address to its data particle */
        super::getParticles()[dtree.low]->octreeNode = (Octree<_DataParticle>*)&dtree;
    } else {
        /* Descend the children of this node. */
	for (int i = 0; i < 8; i++) {
	    if (dtree.child[i]) {
                relinkDataParticlesToOctree(*dtree.child[i]);
	    }//if
	}//for
    }// else
}

template <class _DataParticle, class _Particle>
std::ostream & OctreeInterpolator<_DataParticle,_Particle>::printRadiiOfInfluencePlot(std::ostream & output) {
    output << "set pm3d at s hidden3d 1 ftriangles explicit\n"
           << "set style line 1 lt 3\n"
           << "set grid layerdefault\n"
           << "set palette rgb 9,9,2\n"
           << "unset colorbox \n"
           << "set hidden3d\n"
           << "set parametric\n"
           << "set size square\n"
           << "set samples 36\n"
           << "set isosamples 10,13\n"
           << "set ticslevel 0\n"
           << "unset key\n"
           << "set xlabel 'x'\n"
           << "set ylabel 'y'\n"
           << "set zlabel 'z'\n"
           << "set urange [0:pi]\n"
           << "set vrange [0:2* pi]\n"
           << "splot ";
    _DataParticle ** end = super::getParticles() + super::numberParticles();
    for (_DataParticle ** d = super::getParticles(); d < end; d++ ) {
        output
           << (*d)->radiusOfInfluence << "*sin(u)*cos(v)+" << (*d)->pos[X] << "," 
           << (*d)->radiusOfInfluence << "*sin(u)*sin(v)+" << (*d)->pos[Y] << "," 
           << (*d)->radiusOfInfluence << "*cos(u)+"        << (*d)->pos[Z]
#if 0
           << " w pm3d"
#else
           << " lt 3"
#endif
           << (d<(end-1)?", ":"");

    }
    output << "\npause -1\n";
    return output;
}

template <class _DataParticle>
inline int getChildIndex(const Octree<_DataParticle,InterpolatorFields> * n) THROWS(std::runtime_error) {
    int child;
    for (child = 0; child < 8; child++) {
        if(n->parent->child[child] == n)
            break;
    }

    if (child >= 8) {
        THROW(std::runtime_error,FILE_LINE "Octree node is not a child of its parent?????");
    }

    return child;
}

template <class _DataParticle, class _Particle>
void OctreeInterpolator<_DataParticle,_Particle>::setRadiiOfInfluence(const int nNodes) THROWS(std::runtime_error){
    if (super::numberParticles() == 1) {
        /* take care of the special case when we have only a single data node
         * in the tree. */
        _DataParticle & d = *super::getParticles()[0];
        d.radiusOfInfluence = d.octreeNode->getVolRadius();
        d.nearestNeighbors = 0;
        return;
    }

    _DataParticle ** end = super::getParticles() + super::numberParticles();
    for (_DataParticle ** d = super::getParticles(); d < end; d++) {
        /* First find out which child this octreenode is of its parent.
         */
        if (!(*d)->octreeNode) {
            THROW(std::runtime_error,FILE_LINE "data node has no octree node????");
        } else if (!(*d)->octreeNode->parent) {
            log_severe(FILE_LINE "data node belongs to a root octree node!!?????");
            log_severe(FILE_LINE "cannot set radius of influence for this (%dth) data node",(d-super::getParticles()));
            continue;
        }

        setRadiusOfInfluence(
            *d,
            ((Octree<_DataParticle,InterpolatorFields>*)(*d)->octreeNode)->parent,
            getChildIndex( (Octree<_DataParticle,InterpolatorFields>*)(*d)->octreeNode ),
            nNodes
        );
    }
}

typedef struct {
    int face[3];
    int edge[3];
    int corner;
} NodeNeighbors;

static NodeNeighbors nodeNeighbors[] = {
  /*0:*/{ {1,3,4}, {2,5,7},  6 },
  /*1:*/{ {0,2,5}, {3,4,6},  7 },
  /*2:*/{ {1,3,6}, {0,5,7},  4 },
  /*3:*/{ {0,2,7}, {1,4,6},  5 },
  /*4:*/{ {0,5,7}, {1,3,6},  2 },
  /*5:*/{ {1,4,6}, {0,2,7},  3 },
  /*6:*/{ {2,5,7}, {1,3,4},  0 },
  /*7:*/{ {3,4,6}, {0,2,5},  1 },
};

template <class _DataParticle>
inline int findDimensions(Octree<_DataParticle,InterpolatorFields> * dtree) {
    if (dtree->getDimensions()) return dtree->getDimensions();

    /* otherwise, use dimensions of parent, as a parent is guarenteed to have
     * >= 1 dimensions.
     */
    if (dtree->parent) {
        return dtree->parent->getDimensions();
    }

    /* we have 0-dimensions at the root of the tree? */
    return 0;
}

template <class _DataParticle, class _Particle>
void OctreeInterpolator<_DataParticle,_Particle>::setRadiusOfInfluence(_DataParticle * d,
                                             Octree<_DataParticle,InterpolatorFields> * dtree,
                                             const int child,
                                             const int nNodes ) {
    NodeNeighbors & nbrs = nodeNeighbors[child];
    double dist = 0;
    Octree<_DataParticle,InterpolatorFields> * node = NULL;
    int nearest = 0;

    /* All nodes joined by faces */
    for (int i = 0; i < 3; i++) {
        node = dtree->child[nbrs.face[i]];
        if (!node) continue;
        nearest += node->leaves;
        dist = node->distance(*d);
        if ( dist > d->radiusOfInfluence ) d->radiusOfInfluence = dist;
    }
    /* we'll assume that the density is isotropic (since the tree is supposed
     * to already be partitioned about local center of mass positions.
     * Thus, we double the number of nearest neighbors found, assuming that
     * we'd find the same amount in the other adjacent nodes.
     */
    d->nearestNeighbors += 2 * nearest;
    if (d->nearestNeighbors >= nNodes) goto recurse_upwards;

    /* All nodes joined by edges */
    nearest = 0;
    for (int i = 0; i < 3; i++) {
        node = dtree->child[nbrs.edge[i]];
        if (!node) continue;
        nearest += node->leaves;
        dist = node->distance(*d);
        if ( dist > d->radiusOfInfluence ) d->radiusOfInfluence = dist;
    }
    /* see note above for the factor of 2 */
    d->nearestNeighbors += 2 * nearest;
    if (d->nearestNeighbors >= nNodes) goto recurse_upwards;

    /* Node in the opposite corner */
    node = dtree->child[nbrs.corner];
    if (node) {
        /* see note above for the factor of 2 */
        d->nearestNeighbors += 2 * node->leaves;
        dist = node->distance(*d);
        if ( dist > d->radiusOfInfluence ) d->radiusOfInfluence = dist;
    }

    if (d->nearestNeighbors < nNodes && dtree->parent) {
        /* now call this for this node's parent */
        setRadiusOfInfluence(d,dtree->parent, getChildIndex(dtree), nNodes);
    } else goto recurse_upwards;

    /* set maximum radiusOfInfluence for this Octree node. */
    if (dtree->extraFields.maxRadiusOfInfluence < d->radiusOfInfluence) {
        dtree->extraFields.maxRadiusOfInfluence = d->radiusOfInfluence;
    }
    return;

recurse_upwards:
    /* I'm going to try and rescale the radius of influence to only include
     * nNodes.  This rescaling will work under the assumption that neighbors
     * are isotropic.
     */
    int dimensions = findDimensions(dtree);
    switch(dimensions) {
        case 3:
            /* r  = (V/V_{real})^(1/3) * r_real */
            d->radiusOfInfluence *= fast_pow(double(nNodes)/double(d->nearestNeighbors), 0.33333333333333333333);
            break;
        case 2:
            /* r  = (A/A_{real})^(1/2) * r_real */
            d->radiusOfInfluence *= sqrt(double(nNodes)/double(d->nearestNeighbors));
            break;
        case 1:
            /* r  = (L/L_{real}) * r_real */
            d->radiusOfInfluence *= double(nNodes)/double(d->nearestNeighbors);
            break;
        case 0:
        default:
            THROW(std::runtime_error,FILE_LINE "discovered 0-dimensions for data node?");
    }

    /* set maximum radiusOfInfluence for this Octree node. */
    if (dtree->extraFields.maxRadiusOfInfluence < d->radiusOfInfluence) {
        dtree->extraFields.maxRadiusOfInfluence = d->radiusOfInfluence;
    }

    if (!dtree->parent) return;
    /* pass this radiusOfInfluence up the hierarchy to allow for each
     * level of the tree to have info as to the maximum radius of
     * influence in their sub-trees.
     * */
    for ( Octree<_DataParticle,InterpolatorFields> * node = dtree->parent; node; node = node->parent ) {
        if (node->extraFields.maxRadiusOfInfluence < d->radiusOfInfluence) {
            node->extraFields.maxRadiusOfInfluence = d->radiusOfInfluence;
        }
    }
    /* done */
}

template <class _DataParticle, class _Particle>
void OctreeInterpolator<_DataParticle,_Particle>::gatherInterpolant(const double x[],
                                             const std::vector<int> & dtv,
                                             const Octree<_DataParticle,InterpolatorFields> & node,
                                             std::vector<double> & value_sum,
                                             double & kernel_sum ) const {
    /* if there are 8 or fewer leaves, then it is at least os little or less work to just
     * loop through and gather the interpolant from each of the enclosed data
     * nodes.
     * */
    if ( node.leaves <= 8 ) {
        for (int i = node.low; i <= node.high; i++) {
            _DataParticle & d = *super::getParticles()[i];
            double k = kernel(distance(d,x), d.radiusOfInfluence);
            for (unsigned int j = 0; j < dtv.size(); j++) {
                value_sum[j] += d.data.indx[dtv[j]] * k;
            }/* for */
            kernel_sum += k;
        }/* for */
    } else {
        /* ok, so we must now test (for each child tree) whether we need to
         * descend further. */
        for (int i = 0; i < 8; i++) {
            if ( node.child[i] &&
                 node.child[i]->getVolB().contains(x, 2.0*node.child[i]->extraFields.maxRadiusOfInfluence)
               ) {
                gatherInterpolant(x, dtv, *node.child[i], value_sum, kernel_sum);
            }/* if */
        }/* for */
    }/* else */
}

template <class _DataParticle, class _Particle>
void OctreeInterpolator<_DataParticle,_Particle>::initKernel( double dr_h ) {
    /* We'll try using the spherical spline kernel that Hernquist and Katz so
     * lovingly use in their TREESPH paper:
     * Lars Hernquist and Neal Katz,
     *     ``TREESPH:  A Unification of SPH with the heirarchical tree method,''
     *     The Astrophysical Journal Supplement Series, 70:419-446, 1989 June.
     *
     * W(r,h) = 1/(pi*h^3) * [ ( 1-(3/2)*(r/h).^2 + (3/4)*(r/h).^3 ),   // 0<=r/h<=1
                               ( (1/4)*(2-(r/h)).^3 ),                  // 1<=r/h<=2
                               0                                        // 2<=r/h
                             ];
     *
     * We may as well also include what they suggested in creating a lookup
     * table for this spline.  This should speed up computation a little.
     */
    dr_h_recip = 1/dr_h;
    int nBins = (int) ( 2.0 * dr_h_recip ) + 1;

    /* allocate space for lookup table */
    if (W) delete[] W;
    W = new double[nBins];

    /* Assign lookup table for 0 <= (r/h) <=  1 */
    int i = 0;
    double r_h = 0;
    for (; r_h <= 1.0L; r_h+=dr_h) {
        W[i] = /* M_1_PI */ ( 1.0L - 1.5L*SQR(r_h) + 0.75L*CUBE(r_h) );
        i++;
    }

    /* Assign lookup table for 1 <= (r/h) <=  2 */
    for (double s = 2.0L - r_h; r_h < 2.0L; s-=dr_h,r_h+=dr_h) {
        W[i] = /* M_1_PI */ ( 0.25L * CUBE(s) );
        i++;
    }

    if (i > nBins) {
        log_severe(FILE_LINE "Number of bins assigned (%d) is greater than number allocated (%d), segfault coming!!", i, nBins);
    }

}

} /* namespace octree. */
