// -*- c++ -*-
// $Id: OctreeInterpolator.h,v 1.4 2005/06/07 10:00:12 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: OctreeInterpolator.h,v $
 * Revision 1.4  2005/06/07 10:00:12  olsonse
 * Many changes over the course of time
 *
 * Revision 1.3  2005/01/22 20:27:27  olsonse
 * Added method to clear radius of influence so that it may be changed.
 *
 * Revision 1.2  2005/01/19 11:39:16  olsonse
 * Changed interpolator algorithm to be able to interpolate several data values
 * at the same time.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */

#ifndef OCTREE_INTERP_H
#define OCTREE_INTERP_H

#include "OctreeContainer.h"
#include <vector>

/** \file
 * This is the first documentation for this file as a whole.
 * */

namespace octree {

class InterpolatorFields {
  public:
    inline InterpolatorFields() : maxRadiusOfInfluence(0) {}
    double maxRadiusOfInfluence;
};

/** An interpolation algorithm based on an octree code and field theory
 * garbage.
*/
template <class _DataParticle, class _Particle>
class OctreeInterpolator: public OctreeContainer<_DataParticle,InterpolatorFields> {
    typedef OctreeContainer<_DataParticle,InterpolatorFields> super;
  public:
    /** Default contructor. */
    OctreeInterpolator();

    /** Destructor. */
    ~OctreeInterpolator();

    /** Create and pattern a data octree after the Particle octree.
     * Note that this function will not succeed until the Particle octree has
     * been created.
     *
     * WARNING WARNING WARNING:
     *
     * Make sure that particlesOctree->findRcmAndVol(...) has been run before
     * calling this function.  This function leaves this responsibility to the
     * caller so that speed might be optimized.
     *
     * @throws std::runtime_error
     *     If findRcmAndVol(...) has not yet been called for the Octree.
     */
    template <class PTreeType, class T>
    void createDataOctree(PTreeType * particlesOctree, T & auxillary) THROWS (std::runtime_error);

    /** Create a data Octree from the given DataParticle linked list.
     * The particles are added into this container class.
     *
     * @return false if not successful; true otherwise.
     */
    bool createDataOctree(_DataParticle * data);

  private:
    /** Recursivly create and copy data into the data octree.
     * @see createDataOctree(PTreeType * particlesOctree).
     */
    template <class PTreeType, class T>
    void copyDataFromParticleOctree(Octree<_DataParticle,InterpolatorFields> & dtree,
                                    const PTreeType & ptree,
                                    T & auxillary);

    /** Links each DataParticle with its containing Octree node.
     * @see createDataOctree(_DataParticle * data).
     */
    void relinkDataParticlesToOctree(Octree<_DataParticle,InterpolatorFields> & dtree);

  public:
    /** Initialize the kernel lookup table.
     * @param dr_h
     *     Stepsize (in units of (r/h)) of each bin in the lookup table
     *     (defaults to 0.02, which gives 100 bins).
     * */
    void initKernel( double dr_h = 0.02L );

    /** Return the value of the smoothing kernel.
     * This MUST be called AFTER initKernel() has been called.
     * Otherwise, segmentation violations WILL occur.
     * @param r
     *     Radius at which to evaluate kernel.
     * @param h
     *     Smoothing radius to use for evaluating the kernel.
     * @return
     *     Value of the spherical spline smoothing kernel at (r,h).
     * */
    inline double kernel(double r, double h) const {
        double r_h = r/h;
        if (r_h >= 2.0) return 0;

        int r_h_i = (int)( r_h * dr_h_recip );

        /* return the value of the spherical spline kernel */
        return W[r_h_i] / CUBE(h);
    }

    inline static double distance(const _DataParticle & d, const double x[]) {
        double dx = (d.pos[X] - x[X]),
               dy = (d.pos[Y] - x[Y]),
               dz = (d.pos[Z] - x[Z]);
        return sqrt( SQR(dx) + SQR(dy) + SQR(dz) );
    }

  public:
    /** Set the radius of influence for each data node.
     * @param nNodes
     *     The approximate minimum of nearest neighbor nodes to account for.
     * */
    void setRadiiOfInfluence(int nNodes) THROWS (std::runtime_error);

  private:

    void setRadiusOfInfluence(_DataParticle * d,
                              Octree<_DataParticle,InterpolatorFields> * dtree,
                              const int child,
                              const int nNodes );

  public:
    /** Clear the already set radius of influence.
     */
    inline void clearRadiiOfInfluence() {
        for (int i = super::numberParticles()-1; i >= 0; i--) {
            super::getParticles()[i]->radiusOfInfluence = 0;
        }
    }

    /** Interpolate the given data type dt at x.
     * @param x
     *     Position of the interpolated value.
     * @param dtv
     *     Vector of indices of datatype that is to be interpolated.
     * */
    inline std::vector<double> interpolate(const double x[], const std::vector<int> & dtv) const {
        double denominator = 0;
        std::vector<double> numerator(dtv.size(),0.0);
        gatherInterpolant(x, dtv, *super::getOctree(), numerator, denominator);
        if (denominator == 0) {
            return std::vector<double>();
        } else {
            for (unsigned int i = 0; i < numerator.size(); i++) {
                numerator[i] /= denominator;
            }
            return numerator;
        }
    }

    /** Interpolate the given data type dt at x.
     * @param x
     *     Position of the interpolated value.
     * @param dt
     *     Index of datatype that is to be interpolated.
     * */
    inline double interpolate(const double x[], const int & dt) const {
        std::vector<int> dtv(1,dt);
        std::vector<double> interpolant = interpolate(x,dtv);
        return interpolant.size() != 1 ? 0 : interpolant[0];
    }


  private:
    void   gatherInterpolant(const double x[],
                             const std::vector<int> & dtv,
                             const Octree<_DataParticle,InterpolatorFields> & node,
                             std::vector<double> & value_sum,
                             double & kernel_sum ) const;

  public:
    /** Print a GnuPlot representation of all of the radii of influence.
     * @param output
     *     Output stream to print to.
     * */
    std::ostream & printRadiiOfInfluencePlot(std::ostream & output);

  private:
    double * W; /* smoothing kernel lookup table. */
    double dr_h_recip; /* == 1/ (d (r/h) ). */
};

} /* namespace octree. */

#include "OctreeInterpolator.C"

#endif // OCTREE_INTERP_H

