
#ifndef PTHREADBUILDOCTREE
#define PTHREADBUILDOCTREE



#include <pthread.h>

class PThreadLimiter {
  private:
    pthread_mutex_t active_threads_mutex;
    pthread_mutex_t wait_mutex;
    pthread_cond_t  wait_condition;

    int thread_count;
    int max_threads;

  public:

    PThreadLimiter() {
        pthread_mutex_init(&active_threads_mutex, NULL);
        pthread_mutex_init(&          wait_mutex, NULL);
        pthread_cond_init(&wait_condition, NULL);
        thread_count = 0;

        char * max_thread_str = getenv("NUM_PTHREADS");
        max_threads = 0;

        if ( max_thread_str == NULL ||
             sscanf(max_thread_str, "%d", &max_threads) != 1 ) {

            /* we'll default to one thread only */
            max_threads = 1;
        }
    }

    ~PThreadLimiter() {
        pthread_mutex_destroy(&active_threads_mutex);
        pthread_mutex_destroy(&          wait_mutex);
        pthread_cond_destroy(&wait_condition);
    }


    inline void set_max_pthreads(int mx) {
        max_threads = mx;
    }

    inline const int & get_max_pthreads() const {
        return max_threads;
    }

    /** Releases the given thread number.  */
    inline void releaseThreadNumber() {
        pthread_mutex_lock(&active_threads_mutex);
        thread_count--;
        pthread_mutex_unlock(&active_threads_mutex);

        /* and now, signal to any waiting threads that a new thread can now be
         * run. */
        pthread_cond_signal(&wait_condition);
    }

    /** Acquires a thread number to continue. */
    inline int acquireThreadNumber() {
        int retval = -1;
        pthread_mutex_lock(&active_threads_mutex);
        if ( thread_count < get_max_pthreads() ) {
            retval = thread_count;
            thread_count++;
        }
        pthread_mutex_unlock(&active_threads_mutex);
        return retval;
    }

    inline int waitForAvailableThread() {
        int thread_number = -1;
        pthread_mutex_lock(&wait_mutex);
        while ( (thread_number = acquireThreadNumber()) < 0) {
            pthread_cond_wait(&wait_condition, &wait_mutex);
        }
        pthread_mutex_unlock(&wait_mutex);
        return thread_number;
    }
};

static PThreadLimiter pthreadLimiter;

#define get_max_threads        pthreadLimiter.get_max_pthreads


template < class _Octree>
class PThreadBuildOctree {
  private:
    typedef typename _Octree::OParticle _Particle;
    typedef typename _Octree::OLeafModel _LeafModel;

    class PThreadArgs {
      public:
        _Octree * child;
        _Particle **a;
        _LeafModel leafModel;
        int * leaves;

        PThreadArgs(_Octree * _child = NULL,
                    _Particle *_a[] = NULL,
                    _LeafModel _leafModel = _LeafModel(), int *_leaves = NULL)
        : child(_child), a(_a), leafModel(_leafModel), leaves(_leaves) { }

        PThreadArgs & operator=(const PThreadArgs & that) {
            child = that.child;
            a = that.a;
            leafModel = that.leafModel;
            leaves = that.leaves;
            return *this;
        }
    };


    pthread_attr_t pthread_attr;

    PThreadArgs args[8];
    pthread_t threads[8];
    int retvals[8];


  public:

    PThreadBuildOctree() {
        memset(threads,0,sizeof(threads));
        memset(retvals,0,sizeof(retvals));
        pthread_attr_init(&pthread_attr);
    }

    ~PThreadBuildOctree() {
        pthread_attr_destroy(&pthread_attr);
    }



    static void runBuildOctree(PThreadArgs * _args) {
        PThreadArgs & args = *_args;
        (*args.leaves) = args.child->buildOctree(args.a, args.leafModel);
        pthreadLimiter.releaseThreadNumber();
    }

    inline void buildOctree(const int & i, _Octree * child,
                            _Particle *a[],
                            const _LeafModel & leafModel) {
        /* first wait for a thread to become available */
        pthreadLimiter.waitForAvailableThread();
        args[i] = PThreadArgs(child, a, leafModel, &retvals[i]);
        pthread_create(&threads[i], &pthread_attr, (void*(*)(void*))runBuildOctree, &args[i]);
    };


    inline int gatherLeaves() {
        int leaves = 0;
        /* We wait for each successive child thread and gather. */
        for (int i = 0; i < 8; i++) {
            if (threads[i] != 0) {
                pthread_join(threads[i],NULL);
                leaves += retvals[i];
                threads[i] = 0;
            }
        }

        return leaves;
    }
};



#endif // PTHREADBUILDOCTREE
