// -*- c++ -*-
// $Id: Wall.h,v 1.2 2005/01/19 11:47:44 olsonse Exp $
/*
 * Copyright 2002-2004 Spencer Olson
 *
 * $Log: Wall.h,v $
 * Revision 1.2  2005/01/19 11:47:44  olsonse
 * Added documentation.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */

#ifndef WALL_H
#define WALL_H

#include <stdexcept>
#include <olson-tools/Vector.h>
#include <olson-tools/physical.h>
#include <olson-tools/random/random.h>
#include <olson-tools/Distribution.h>
#include "FlatSurface.h"

class Wall : public virtual FlatSurface {
  public:
    /** The super class of Wall is a FlatSurface. */
    typedef FlatSurface super;

    /** Wall constructor.
     * @param normal
     *     Normal to the surface of the wall.
     * @param point
     *     A point lying on the wall somewhere.
     * @param velocity
     *     Velocity of Wall (this is useful for Couette flows).
     * @param temp
     *     Temperature of the wall for consideration when performing diffusive
     *     reflections [Defaults to room temperature].
     * @param rough
     *     Set the surface roughness (0 to 1);  roughness is the probability
     *     that a particle will be reflected diffusely from the wall [Defaults
     *     to -1, or specular reflecting only].
     * @param allow_movement
     *     Specify which directions are allowed to change under the reflection
     *     from the wall.  This is useful for 1-D, 2-D simulations.
     * @param number_distrib_bins
     *     Set the number of bins to use in inverting the temperature
     *     distributions of the Wall (gaussian and maxwellian).
     * @see setSurface.
     */
    inline Wall(const Vector<double,3> & normal = V3(1.,0.,0.),
                const Vector<double,3> & point = V3(0.,0.,0.),
                const Vector<double,3> & velocity = V3(0.,0.,0.),
                const double & temp = 300 * physical::unit::Kelvin,
                const double & rough = -1,
                const Vector<bool,3> & allow_movement = true,
                const int & number_distrib_bins = 1000 ) :
                super(normal,point,allow_movement), distrib_length(0), distrib(NULL), nbin_distrib(0) {
        setSurface(normal,point,velocity,temp,rough,allow_movement,number_distrib_bins);
    }

    inline void setSurface(const Vector<double,3> & normal,
                           const Vector<double,3> & point,
                           const Vector<double,3> & velocity = V3(0.,0.,0.),
                           const double & temp = 300 * physical::unit::Kelvin,
                           const double & rough = -1,
                           const Vector<bool,3> & allow_movement = true,
                           const int & number_distrib_bins = 1000 ) {
        /* demolish old Wall/s first. */
        bulldoze();

        super::setSurface(normal,point,allow_movement);
        setSurfaceMatrices();
        temperature = temp;
        roughness = rough;
        v_wall = velocity;
        total_specular_reflections = 0;
        total_diffusive_reflections = 0;
        distrib_length = 0;
        distrib = NULL;
        nbin_distrib = number_distrib_bins;
    }

    /** Destructor.  bulldoze Wall. */
    inline ~Wall() { bulldoze(); }

  private:
    /** Wall bulldozing. */
    inline void bulldoze() {
        if (distrib_length>0) {
            for (int i = 0; i < distrib_length; i++) {
                if (distrib[i]) delete distrib[i];
            }

            /* distrib is allocated with realloc. */
            free(distrib);
            distrib = NULL;
            distrib_length = 0;
        }
    }

  public:

    enum REFLECTION_TYPE {
        NO_REFLECTION=0,
        DIFFUSE_REFLECTION=1,
        SPECULAR_REFLECTION=2
    };

    /** Reflect a particles.
     * Note:  This code will not correctly reflect a Particle where the Wall
     * is in the vicinity of a strong field gradient.  This code assumes the
     * velocity for both particle_old and particle are the same (it only looks
     * at velocity of particle).
     *
     * Note that this code may not work for potentials where the particle
     * makes very quick loop-de-loops near the surface of the wall.  In this
     * case, the timestep would have to be decreased in order to make this
     * test more often.  
     *
     * @param particle_old
     *     (input) Copy of Particle before traversal through Wall; (output)
     *     Location of reflection (wall intersection point) with the new
     *     velocity IF the particle was reflected. 
     *
     * @param particle
     *     Particle to reflect.
     *
     * @param volb
     *     Bounding volume of wall to determine its extents.
     *     A reflection only occurs when the intersection with the wall is
     *     contained by this bounding volume.
     *
     * @returns reflection type (specular, diffuse, or none).
     */
    template <class _Particle, class _Container>
    inline enum REFLECTION_TYPE reflect(_Particle & particle_old,
                                        _Particle & particle,
                                        const _Container & volb ) {
        Vector<double,3> & v = V3C(particle.vel);
        Vector<double,3> & r = V3C(particle.pos);
        /* And now do an active rotation for reflecting the particles
         * position.  We first must determine the point of intersection of the
         * particle with the wall.
         * for:
         *      t :  time of intersection
         *      r :  position of particle
         *      v :  velocity of particle
         *      p :  a point on the reflecting wall
         *      n :  normal to reflecting wall
         *
         *      t = \frac{ n \cdot (p - r) } { N \cdot v }
         */
        double side_sign = 0;
        Vector<double,3> iloc(v);
        double t = intersectTime(particle_old,particle,volb,iloc,side_sign);

        if ( t > 0 ) {
            /* This reflection has not yet happened. */
            return NO_REFLECTION;
        }


        /* ok, finally ready to continue... */
        enum REFLECTION_TYPE retval;
        if (MTRNGrand() > roughness) {
            /* and specular coordinate reflection gives: */
            v = M_r * v;

            /* FIXME:  account for wall velocity with specular reflections!
             * This is important for walls that do not move parallel to their
             * tangents.  Perhaps we can do something like adding the
             * transverse component of the wall velocity.  This would
             * necessitate rotating the coordinate system as is done in the
             * case of diffuse reflections (below).
             */

            /* increment our reflection counter */
            total_specular_reflections++;

            retval = SPECULAR_REFLECTION;
        } else {
            /* perform diffusive reflection from a hot wall.
             * first get a velocity in reference frame of hot wall from
             * sin(2\theta) distribution.
             */
            /* get the speed from a maxwellian distribution. */
            if (distrib_length <= particle.type || distrib[particle.type] == NULL) {
                addDistrib(particle.type,particle.mass());
            }

#if 0
            {
                /* this is code that bird uses.  It seems to give the same
                 * results as mine. Mine is faster. */
                static double betaw = sqrt(2.0 * physical::constant::K_B * temperature / particle_info[particle.type].mass );

                double ran = max(1.e-8,MTRNGrand());
                double B   = sqrt(-log(ran));

                double A   = 2.0 * physical::constant::pi * MTRNGrand();
                v[X] = B * sin(A) * betaw;
                v[Y] = B * cos(A) * betaw;

                ran = max(1.e-8, MTRNGrand());
                v[Z]  = sqrt(-log(ran)) * betaw;
            }
#else
            /* my code */
            v[X] = distrib[particle.type]->gaussian();
            v[Y] = distrib[particle.type]->gaussian();
            v[Z] = distrib[particle.type]->maxwellian() * side_sign;
#endif

            /* and rotate into the correct coordinate system. */
            v = R * v;

            /* and now translate into the lab frame using a Galilaen
             * transformation. */
            v += v_wall;

            total_diffusive_reflections++;

            retval = DIFFUSE_REFLECTION;
        }

        /* set the new position and velocity.
         * The following is equivalent to : r = r0 + v * t, where r0 is the
         * position of the intersection and t is the (positive) amount of time
         * since the reflection. */
        r = v; r.compMult(super::allowed_movement); r *= -t; r += iloc;

        /* now update the old particle position and velocity to be the
         * reflection position and the reflected velocity. */
        V3C(particle_old.vel) = v;
        V3C(particle_old.pos) = iloc;

        return retval;

    }/* reflect */

  private:
    /** Initialize the class for work.
     * This must be done AFTER the parameters are chosen, and
     * BEFORE any atoms can be inserted into the module.
     * Note that a velocity component of the wall normal to the wall may be a
     * bit problematic.
     * */
    void setSurfaceMatrices();

    /** The temperature of the wall.
     * Temperature only affects diffuse reflections.
     */
    double temperature;

    void addDistrib( const int & particle_type, const double & particle_mass );

  public:
    /** This is a number between 0 and 1 to indicate how probable it is that a
     * diffuse reflection with the wall should take place.
     */
    double roughness;

    /** The wall velocity. */
    Vector<double,3> v_wall;

    inline const unsigned long long & totalSpecularRefl() const {
        return total_specular_reflections;
    }

    inline const unsigned long long & totalDiffuseRefl() const {
        return total_diffusive_reflections;
    }

  private:
    SquareMatrix<double,3> M_r; /* for performing specular reflection. */
    SquareMatrix<double,3> R;   /* for rotating diffuse reflection back to lab frame. */
    unsigned long long total_specular_reflections, total_diffusive_reflections;

    class wall_distrib {
      private:
        double beta;
        double sigma_v;
      public:
        inline wall_distrib(const double & temperature,
                            const int & particle_type,
                            const double & particle_mass,
                            const int & nbins);

        Distribution maxwellian;
        Distribution gaussian;
    };

    int distrib_length;
    wall_distrib ** distrib;
    int nbin_distrib;
};

#endif // WALL_H
